from __future__ import annotations

import json
import logging
import os
import re
import threading
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Dict, Optional, Tuple

from abstractruntime import Runtime, WorkflowRegistry, WorkflowSpec, persist_workflow_snapshot
from abstractruntime.visualflow_compiler import compile_visualflow
from abstractruntime.workflow_bundle import WorkflowBundle, WorkflowBundleError, open_workflow_bundle

from ..memory_store import build_gateway_memory_embedder, open_gateway_memory_store
from ..provider_defaults import ProviderModelConfigError, resolve_gateway_provider_model
from ..workflow_deprecations import WorkflowDeprecatedError, WorkflowDeprecationStore


logger = logging.getLogger(__name__)


def _namespace(bundle_id: str, flow_id: str) -> str:
    return f"{bundle_id}:{flow_id}"


def _bundle_ref(bundle_id: str, bundle_version: str) -> str:
    bid = str(bundle_id or "").strip()
    bv = str(bundle_version or "").strip()
    if not bid:
        raise ValueError("bundle_id is required")
    if not bv:
        raise ValueError("bundle_version is required")
    return f"{bid}@{bv}"


def _split_bundle_ref(raw: str) -> tuple[str, Optional[str]]:
    s = str(raw or "").strip()
    if not s:
        return ("", None)
    if "@" not in s:
        return (s, None)
    a, b = s.split("@", 1)
    a = a.strip()
    b = b.strip()
    if not a:
        return ("", None)
    if not b:
        return (a, None)
    return (a, b)


def _try_parse_semver(v: str) -> Optional[tuple[int, int, int]]:
    s = str(v or "").strip()
    if not s:
        return None
    parts = [p.strip() for p in s.split(".")]
    if not parts or any(not p for p in parts):
        return None
    nums: list[int] = []
    for p in parts:
        if not p.isdigit():
            return None
        nums.append(int(p))
    while len(nums) < 3:
        nums.append(0)
    return (nums[0], nums[1], nums[2])


def _pick_latest_version(versions: Dict[str, WorkflowBundle]) -> str:
    items = [(str(k), v) for k, v in (versions or {}).items() if isinstance(k, str)]
    if not items:
        return "0.0.0"

    if all(_try_parse_semver(ver) is not None for ver, _ in items):
        return max(items, key=lambda x: _try_parse_semver(x[0]) or (0, 0, 0))[0]

    # Fallback when versions are not semver-like: prefer newest created_at.
    def _key(x: tuple[str, WorkflowBundle]) -> tuple[str, str]:
        ver, b = x
        created = str(getattr(getattr(b, "manifest", None), "created_at", "") or "")
        return (created, ver)

    return max(items, key=_key)[0]


def _coerce_namespaced_id(*, bundle_id: Optional[str], flow_id: str, default_bundle_id: Optional[str]) -> str:
    fid = str(flow_id or "").strip()
    if not fid:
        raise ValueError("flow_id is required")

    bid = str(bundle_id or "").strip() if isinstance(bundle_id, str) else ""
    if bid:
        # If the caller passed a fully-qualified id (bundle:flow), allow it as-is so
        # clients can safely send both {bundle_id, flow_id} without producing a
        # double-namespace like "bundle:bundle:flow".
        if ":" in fid:
            prefix = fid.split(":", 1)[0].strip()
            if prefix == bid:
                return fid
            raise ValueError(
                f"flow_id '{fid}' is already namespaced, but bundle_id '{bid}' was also provided; "
                "omit bundle_id or pass a non-namespaced flow_id"
            )
        return _namespace(bid, fid)

    # Allow passing a fully-qualified id as flow_id.
    if ":" in fid:
        return fid

    if default_bundle_id:
        return _namespace(default_bundle_id, fid)

    raise ValueError("bundle_id is required when multiple bundles are loaded (or pass flow_id as 'bundle:flow')")


def _namespace_visualflow_raw(
    *,
    raw: Dict[str, Any],
    bundle_id: str,
    flow_id: str,
    id_map: Dict[str, str],
) -> Dict[str, Any]:
    """Return a namespaced copy of VisualFlow JSON, rewriting internal subflow references."""
    fid = str(flow_id or "").strip()
    if not fid:
        raise ValueError("flow_id is required")

    namespaced_id = id_map.get(fid) or _namespace(bundle_id, fid)

    def _maybe_rewrite(v: Any) -> Any:
        if isinstance(v, str):
            s = v.strip()
            if s in id_map:
                return id_map[s]
        return v

    out: Dict[str, Any] = dict(raw)
    out["id"] = namespaced_id

    # Copy/normalize nodes to avoid mutating the original object and to ensure nested dicts
    # are not shared by reference.
    nodes_raw = out.get("nodes")
    if isinstance(nodes_raw, list):
        new_nodes: list[Any] = []
        try:
            from abstractruntime.visualflow_compiler.visual.agent_ids import visual_react_workflow_id
        except Exception:  # pragma: no cover
            visual_react_workflow_id = None  # type: ignore[assignment]

        for n_any in nodes_raw:
            n = n_any if isinstance(n_any, dict) else None
            if n is None:
                new_nodes.append(n_any)
                continue

            n2: Dict[str, Any] = dict(n)
            node_type = n2.get("type")
            type_str = node_type.value if hasattr(node_type, "value") else str(node_type or "")
            data0 = n2.get("data")
            data = dict(data0) if isinstance(data0, dict) else {}

            if type_str == "subflow":
                for key in ("subflowId", "flowId", "workflowId", "workflow_id"):
                    if key in data:
                        data[key] = _maybe_rewrite(data.get(key))

            if type_str == "agent":
                cfg0 = data.get("agentConfig")
                cfg = dict(cfg0) if isinstance(cfg0, dict) else {}
                node_id = str(n2.get("id") or "").strip()
                if node_id and callable(visual_react_workflow_id):
                    cfg["_react_workflow_id"] = visual_react_workflow_id(flow_id=namespaced_id, node_id=node_id)
                if cfg:
                    data["agentConfig"] = cfg

            n2["data"] = data
            new_nodes.append(n2)

        out["nodes"] = new_nodes

    # Shallow-copy edges for consistency (not strictly required).
    edges_raw = out.get("edges")
    if isinstance(edges_raw, list):
        out["edges"] = [dict(e) if isinstance(e, dict) else e for e in edges_raw]

    return out


def _env(name: str, fallback: Optional[str] = None) -> Optional[str]:
    v = os.getenv(name)
    if v is not None and str(v).strip():
        return v
    if fallback:
        v2 = os.getenv(fallback)
        if v2 is not None and str(v2).strip():
            return v2
    return None


def _bool_text(raw: Any) -> Optional[bool]:
    if raw is None:
        return None
    text = str(raw).strip().lower()
    if text in {"1", "true", "yes", "y", "on"}:
        return True
    if text in {"0", "false", "no", "n", "off"}:
        return False
    return None


def _int_text(raw: Any) -> Optional[int]:
    if raw is None:
        return None
    try:
        value = int(str(raw).strip())
    except Exception:
        return None
    return value if value >= 0 else None


def _ensure_runtime_namespace(vars_obj: Dict[str, Any]) -> Dict[str, Any]:
    rt_ns = vars_obj.get("_runtime")
    if not isinstance(rt_ns, dict):
        rt_ns = {}
        vars_obj["_runtime"] = rt_ns
    return rt_ns


def _node_type_from_raw(n: Any) -> str:
    if not isinstance(n, dict):
        return ""
    t = n.get("type")
    return t.value if hasattr(t, "value") else str(t or "")


def _scan_flows_for_llm_defaults(flows_by_id: Dict[str, Dict[str, Any]]) -> Optional[Tuple[str, str]]:
    """Return a best-effort (provider, model) pair from VisualFlow node configs."""
    def _pair_from_mapping(raw: Any) -> Optional[Tuple[str, str]]:
        if not isinstance(raw, dict):
            return None
        provider = raw.get("provider")
        model = raw.get("model")
        if isinstance(provider, str) and provider.strip() and isinstance(model, str) and model.strip():
            return (provider.strip().lower(), model.strip())
        return None

    for raw in (flows_by_id or {}).values():
        nodes = raw.get("nodes")
        if not isinstance(nodes, list):
            continue
        for n in nodes:
            t = _node_type_from_raw(n)
            data = n.get("data") if isinstance(n, dict) else None
            if not isinstance(data, dict):
                data = {}

            if t == "llm_call":
                cfg = data.get("effectConfig")
                pair = _pair_from_mapping(cfg)
                if pair is not None:
                    return pair

            if t == "agent":
                cfg = data.get("agentConfig")
                pair = _pair_from_mapping(cfg)
                if pair is not None:
                    return pair

            pair = _pair_from_mapping(data.get("pinDefaults"))
            if pair is not None:
                return pair

    return None


def _flow_uses_llm(raw: Dict[str, Any]) -> bool:
    nodes = raw.get("nodes")
    if not isinstance(nodes, list):
        return False
    for n in nodes:
        t = _node_type_from_raw(n)
        if t in {"llm_call", "agent"}:
            return True
    return False


def _flow_uses_tools(raw: Dict[str, Any]) -> bool:
    nodes = raw.get("nodes")
    if not isinstance(nodes, list):
        return False
    for n in nodes:
        t = _node_type_from_raw(n)
        if t in {"tool_calls", "agent"}:
            return True
    return False


def _flow_uses_memory_kg(raw: Dict[str, Any]) -> bool:
    nodes = raw.get("nodes")
    if not isinstance(nodes, list):
        return False
    for n in nodes:
        t = _node_type_from_raw(n)
        if t in {"memory_kg_assert", "memory_kg_query"}:
            return True
    return False


def _collect_agent_nodes(raw: Dict[str, Any]) -> list[tuple[str, Dict[str, Any]]]:
    nodes = raw.get("nodes")
    if not isinstance(nodes, list):
        return []
    out: list[tuple[str, Dict[str, Any]]] = []
    for n in nodes:
        if _node_type_from_raw(n) != "agent":
            continue
        if not isinstance(n, dict):
            continue
        node_id = str(n.get("id") or "").strip()
        if not node_id:
            continue
        data = n.get("data")
        data = data if isinstance(data, dict) else {}
        cfg0 = data.get("agentConfig")
        cfg = dict(cfg0) if isinstance(cfg0, dict) else {}
        out.append((node_id, cfg))
    return out


def _visual_event_listener_workflow_id(*, flow_id: str, node_id: str) -> str:
    # Local copy of the canonical id scheme (kept simple and deterministic).
    import re

    safe_re = re.compile(r"[^a-zA-Z0-9_-]+")

    def _sanitize(v: str) -> str:
        s = str(v or "").strip()
        if not s:
            return "unknown"
        s = safe_re.sub("_", s)
        return s or "unknown"

    return f"visual_event_listener_{_sanitize(flow_id)}_{_sanitize(node_id)}"


@dataclass
class WorkflowBundleGatewayHost:
    """Gateway host that starts/ticks runs from WorkflowBundles (no AbstractFlow import).

    Compiles `manifest.flows` (VisualFlow JSON) via AbstractRuntime's VisualFlow compiler
    (single semantics).
    """

    bundles_dir: Path
    data_dir: Path
    dynamic_flows_dir: Path
    deprecation_store: WorkflowDeprecationStore
    # bundle_id -> bundle_version -> WorkflowBundle
    bundles: Dict[str, Dict[str, WorkflowBundle]]
    # bundle_id -> latest bundle_version
    latest_bundle_versions: Dict[str, str]
    runtime: Runtime
    workflow_registry: WorkflowRegistry
    specs: Dict[str, WorkflowSpec]
    event_listener_specs_by_root: Dict[str, list[str]]
    _default_bundle_id: Optional[str]
    memory_store: Optional[Any] = None
    memory_store_info: Optional[Dict[str, Any]] = None
    _lock: Any = field(default_factory=threading.RLock, repr=False, compare=False)

    @staticmethod
    def _dynamic_flow_filename(workflow_id: str) -> str:
        safe_re = re.compile(r"[^a-zA-Z0-9._-]+")
        s = safe_re.sub("_", str(workflow_id or "").strip())
        if not s or s in {".", ".."}:
            s = "workflow"
        return f"{s}.json"

    def _dynamic_flow_path(self, workflow_id: str) -> Path:
        return Path(self.dynamic_flows_dir) / self._dynamic_flow_filename(workflow_id)

    def load_dynamic_visualflow(self, workflow_id: str) -> Optional[Dict[str, Any]]:
        """Load a persisted dynamic VisualFlow JSON from disk (best-effort)."""
        wid = str(workflow_id or "").strip()
        if not wid:
            return None
        p = self._dynamic_flow_path(wid)
        if not p.exists() or not p.is_file():
            return None
        try:
            raw = json.loads(p.read_text(encoding="utf-8"))
        except Exception:
            return None
        return raw if isinstance(raw, dict) else None

    def register_dynamic_visualflow(self, raw: Dict[str, Any], *, persist: bool = True) -> str:
        """Register a VisualFlow JSON object as a dynamic workflow (durable in data_dir).

        This is used for gateway-generated wrapper workflows (e.g. scheduled runs).
        """
        if not isinstance(raw, dict):
            raise TypeError("Dynamic VisualFlow must be an object")
        wid = str(raw.get("id") or "").strip()
        if not wid:
            raise ValueError("Dynamic VisualFlow missing required 'id'")

        spec = compile_visualflow(raw)
        self.workflow_registry.register(spec)
        self.specs[str(spec.workflow_id)] = spec

        if persist:
            try:
                Path(self.dynamic_flows_dir).mkdir(parents=True, exist_ok=True)
                p = self._dynamic_flow_path(str(spec.workflow_id))
                p.write_text(json.dumps(raw, ensure_ascii=False, indent=2), encoding="utf-8")
            except Exception as e:
                raise RuntimeError(f"Failed to persist dynamic VisualFlow: {e}") from e

        return str(spec.workflow_id)

    def upsert_dynamic_visualflow(self, raw: Dict[str, Any], *, persist: bool = True) -> str:
        """Register or replace a dynamic VisualFlow JSON object (durable in data_dir).

        This is used for gateway-generated wrapper workflows that must be edited in-place
        (e.g. rescheduling a recurrent job without killing the parent run).
        """
        if not isinstance(raw, dict):
            raise TypeError("Dynamic VisualFlow must be an object")
        wid = str(raw.get("id") or "").strip()
        if not wid:
            raise ValueError("Dynamic VisualFlow missing required 'id'")

        spec = compile_visualflow(raw)
        with self._lock:
            try:
                existing = self.workflow_registry.get(spec.workflow_id)
            except Exception:
                existing = None
            if existing is not None:
                try:
                    self.workflow_registry.unregister(spec.workflow_id)
                except Exception:
                    pass
            try:
                self.workflow_registry.register(spec)
            except Exception as e:
                raise RuntimeError(f"Failed to register workflow '{spec.workflow_id}': {e}") from e

            self.specs[str(spec.workflow_id)] = spec

            if persist:
                try:
                    Path(self.dynamic_flows_dir).mkdir(parents=True, exist_ok=True)
                    p = self._dynamic_flow_path(str(spec.workflow_id))
                    p.write_text(json.dumps(raw, ensure_ascii=False, indent=2), encoding="utf-8")
                except Exception as e:
                    raise RuntimeError(f"Failed to persist dynamic VisualFlow: {e}") from e

        return str(spec.workflow_id)

    def _try_register_dynamic_workflow_from_disk(self, workflow_id: str) -> Optional[WorkflowSpec]:
        wid = str(workflow_id or "").strip()
        if not wid:
            return None
        p = self._dynamic_flow_path(wid)
        if not p.exists() or not p.is_file():
            return None
        try:
            raw = json.loads(p.read_text(encoding="utf-8"))
        except Exception:
            return None
        if not isinstance(raw, dict):
            return None
        try:
            spec = compile_visualflow(raw)
        except Exception:
            return None
        try:
            self.workflow_registry.register(spec)
            self.specs[str(spec.workflow_id)] = spec
        except Exception:
            return None
        return spec

    @staticmethod
    def load_from_dir(
        *,
        bundles_dir: Path,
        data_dir: Path,
        run_store: Any,
        ledger_store: Any,
        artifact_store: Any,
    ) -> "WorkflowBundleGatewayHost":
        base = Path(bundles_dir).expanduser().resolve()
        if not base.exists():
            if str(base.name or "").lower().endswith(".flow"):
                raise FileNotFoundError(f"bundles_dir file does not exist: {base}")
            try:
                base.mkdir(parents=True, exist_ok=True)
                logger.warning("bundles_dir did not exist; created %s", base)
            except Exception as e:
                raise FileNotFoundError(f"bundles_dir does not exist and could not be created: {base} ({e})") from e

        bundle_paths: list[Path] = []
        if base.is_file():
            bundle_paths = [base]
        else:
            bundle_paths = sorted([p for p in base.glob("*.flow") if p.is_file()])

        bundles_by_id: Dict[str, Dict[str, WorkflowBundle]] = {}
        for p in bundle_paths:
            try:
                b = open_workflow_bundle(p)
                bid = str(getattr(getattr(b, "manifest", None), "bundle_id", "") or "").strip()
                bver = str(getattr(getattr(b, "manifest", None), "bundle_version", "0.0.0") or "0.0.0").strip() or "0.0.0"
                if not bid:
                    raise WorkflowBundleError(f"Bundle '{p}' has empty bundle_id")
                versions = bundles_by_id.setdefault(bid, {})
                if bver in versions:
                    logger.warning("Duplicate bundle version '%s@%s' at %s; keeping first", bid, bver, p)
                    continue
                versions[bver] = b
            except Exception as e:
                logger.warning("Failed to load bundle %s: %s", p, e)

        if not bundles_by_id:
            logger.warning("No bundles found in %s (expected *.flow). Starting gateway with zero loaded bundles.", base)

        default_bundle_id = next(iter(bundles_by_id.keys())) if len(bundles_by_id) == 1 else None
        latest_versions: Dict[str, str] = {bid: _pick_latest_version(versions) for bid, versions in bundles_by_id.items()}

        data_root = Path(data_dir).expanduser().resolve()
        dep_store = WorkflowDeprecationStore(path=data_root / "workflow_deprecations.json")
        dynamic_dir = data_root / "dynamic_flows"
        try:
            dynamic_dir.mkdir(parents=True, exist_ok=True)
        except Exception:
            # Best-effort: dynamic workflows are optional.
            pass

        # Build runtime + registry and register all workflow specs.
        wf_reg: WorkflowRegistry = WorkflowRegistry()
        specs: Dict[str, WorkflowSpec] = {}
        flows_by_namespaced_id: Dict[str, Dict[str, Any]] = {}

        for bid, versions in bundles_by_id.items():
            for bver, b in versions.items():
                bundle_ref = _bundle_ref(bid, bver)
                man = b.manifest
                if not man.flows:
                    raise WorkflowBundleError(f"Bundle '{bid}@{bver}' has no flows (manifest.flows is empty)")

                flow_ids = set(man.flows.keys())
                id_map = {flow_id: _namespace(bundle_ref, flow_id) for flow_id in flow_ids}

                for flow_id, rel in man.flows.items():
                    raw = b.read_json(rel)
                    if not isinstance(raw, dict):
                        raise WorkflowBundleError(f"VisualFlow JSON for '{bid}@{bver}:{flow_id}' must be an object")
                    namespaced_raw = _namespace_visualflow_raw(
                        raw=raw,
                        bundle_id=bundle_ref,
                        flow_id=flow_id,
                        id_map=id_map,
                    )
                    flows_by_namespaced_id[str(namespaced_raw.get("id") or _namespace(bundle_ref, flow_id))] = namespaced_raw
                    try:
                        spec = compile_visualflow(namespaced_raw)
                    except Exception as e:
                        raise WorkflowBundleError(f"Failed compiling VisualFlow '{bid}@{bver}:{flow_id}': {e}") from e
                    wf_reg.register(spec)
                    specs[str(spec.workflow_id)] = spec

        # Load dynamic flows persisted in data_dir (e.g. scheduled wrapper flows).
        try:
            for p in sorted(dynamic_dir.glob("*.json")):
                if not p.is_file():
                    continue
                try:
                    raw = json.loads(p.read_text(encoding="utf-8"))
                except Exception as e:
                    logger.warning("Failed to read dynamic flow %s: %s", p, e)
                    continue
                if not isinstance(raw, dict):
                    continue
                try:
                    spec = compile_visualflow(raw)
                except Exception as e:
                    logger.warning("Failed compiling dynamic flow %s: %s", p, e)
                    continue
                try:
                    wf_reg.register(spec)
                    specs[str(spec.workflow_id)] = spec
                except Exception as e:
                    logger.warning("Failed registering dynamic flow %s: %s", p, e)
                    continue
        except Exception:
            pass

        needs_llm = any(_flow_uses_llm(raw) for raw in flows_by_namespaced_id.values())
        needs_tools = any(_flow_uses_tools(raw) for raw in flows_by_namespaced_id.values())
        needs_memory_kg = any(_flow_uses_memory_kg(raw) for raw in flows_by_namespaced_id.values())

        extra_effect_handlers: Dict[Any, Any] = {}
        memory_store_obj: Optional[Any] = None
        memory_store_info: Optional[Dict[str, Any]] = None
        if needs_memory_kg:
            try:
                from abstractruntime.integrations.abstractmemory.effect_handlers import build_memory_kg_effect_handlers
                from abstractruntime.storage.artifacts import utc_now_iso
            except Exception as e:  # pragma: no cover
                raise WorkflowBundleError(
                    "Bundle uses memory_kg_* nodes but AbstractMemory integration is not available. "
                    "Install/repair with `pip install abstractgateway`."
                ) from e

            embedder = build_gateway_memory_embedder(base_dir=Path(data_root))

            try:
                memory_resolution = open_gateway_memory_store(base_dir=Path(data_root), embedder=embedder)
                memory_store_obj = memory_resolution.store
                memory_store_info = memory_resolution.public_dict()
            except Exception as e:
                raise WorkflowBundleError(
                    "Bundle uses memory_kg_* nodes, but Gateway could not open the configured AbstractMemory store. "
                    f"{e}"
                ) from e
            for warning in list((memory_store_info or {}).get("warnings") or []):
                logger.warning("Gateway memory store warning: %s", warning)

            extra_effect_handlers = build_memory_kg_effect_handlers(store=memory_store_obj, run_store=run_store, now_iso=utc_now_iso)

        # Optional AbstractCore integration for LLM_CALL + TOOL_CALLS.
        if needs_llm or needs_tools:
            try:
                from abstractruntime.integrations.abstractcore.default_tools import build_default_tool_map
                from abstractruntime.integrations.abstractcore.tool_executor import (
                    AbstractCoreToolExecutor,
                    MappingToolExecutor,
                    PassthroughToolExecutor,
                )
            except Exception as e:  # pragma: no cover
                raise WorkflowBundleError(
                    "This bundle requires LLM/tool execution, but AbstractRuntime was installed "
                    "without AbstractCore integration. Install `abstractruntime[abstractcore]` "
                    "(and ensure `abstractcore` is importable)."
                ) from e

            # Tool execution policy:
            # - approval (default): execute safe tools locally; require explicit approval for dangerous/unknown tools.
            # - passthrough: require explicit approval for *all* tools (then execute in-process on resume).
            # - delegated: do not execute tools; TOOL_CALLS yields a durable JOB wait for external executors.
            # - local/local_all: execute all tools locally (dev-only; unsafe).
            tool_mode = str(_env("ABSTRACTGATEWAY_TOOL_MODE") or "approval").strip().lower()
            # Always build a concrete in-process executor so thin-client approvals can execute tools
            # inside the runtime (no bridge-owned tool execution).
            base_executor: Any = MappingToolExecutor(build_default_tool_map())
            if tool_mode in {"local", "local_all"}:
                tool_executor = base_executor
            elif tool_mode in {"approval", "local_approval", "local-approval"}:
                try:
                    from abstractruntime.integrations.abstractcore.tool_executor import ApprovalToolExecutor, ToolApprovalPolicy

                    tool_executor = ApprovalToolExecutor(delegate=base_executor, policy=ToolApprovalPolicy())
                except Exception:
                    tool_executor = base_executor
            elif tool_mode in {"delegated", "delegate", "job"}:
                tool_executor = PassthroughToolExecutor(mode="delegated")
            else:
                # Back-compat: "passthrough" means "approval required for all tools".
                try:
                    from abstractruntime.integrations.abstractcore.tool_executor import ApprovalToolExecutor, ToolApprovalPolicy

                    tool_executor = ApprovalToolExecutor(delegate=base_executor, policy=ToolApprovalPolicy(auto_approve_tools=set()))
                except Exception:
                    tool_executor = PassthroughToolExecutor(mode="approval_required")

            if needs_llm:
                try:
                    from abstractruntime.integrations.abstractcore.factory import create_local_runtime
                except Exception as e:  # pragma: no cover
                    raise WorkflowBundleError(
                        "LLM nodes require AbstractRuntime AbstractCore integration. "
                        "Install `abstractruntime[abstractcore]`."
                    ) from e

                try:
                    provider, model = resolve_gateway_provider_model(
                        flow_defaults=_scan_flows_for_llm_defaults(flows_by_namespaced_id),
                        purpose="bundle LLM execution",
                    ).require()
                except ProviderModelConfigError as e:
                    raise WorkflowBundleError(
                        "Bundle contains LLM nodes but no default provider/model is configured. "
                        "Set ABSTRACTGATEWAY_PROVIDER and ABSTRACTGATEWAY_MODEL, configure defaults via "
                        "`abstractcore --config`, or ensure the flow JSON includes provider/model on at least "
                        "one llm_call/agent node."
                    ) from e

                runtime = create_local_runtime(
                    provider=provider,
                    model=model,
                    run_store=run_store,
                    ledger_store=ledger_store,
                    artifact_store=artifact_store,
                    tool_executor=tool_executor,
                    extra_effect_handlers=extra_effect_handlers,
                )
                runtime.set_workflow_registry(wf_reg)
            else:
                # Tools-only runtime: avoid constructing an LLM client.
                from abstractruntime.core.models import EffectType
                from abstractruntime.integrations.abstractcore.effect_handlers import make_tool_calls_handler

                runtime = Runtime(
                    run_store=run_store,
                    ledger_store=ledger_store,
                    workflow_registry=wf_reg,
                    artifact_store=artifact_store,
                    effect_handlers={
                        EffectType.TOOL_CALLS: make_tool_calls_handler(
                            tools=tool_executor,
                            artifact_store=artifact_store,
                            run_store=run_store,
                        ),
                        **extra_effect_handlers,
                    },
                )
                try:  # pragma: no cover
                    setter = getattr(runtime, "set_tool_executor_for_resume", None)
                    if callable(setter):
                        setter(tool_executor)
                except Exception:
                    pass
        else:
            runtime = Runtime(
                run_store=run_store,
                ledger_store=ledger_store,
                workflow_registry=wf_reg,
                artifact_store=artifact_store,
                effect_handlers=extra_effect_handlers,
            )

        # Register derived workflows required by VisualFlow semantics:
        # - per-Agent-node ReAct subworkflows
        # - per-OnEvent-node listener workflows (Blueprint-style)
        event_listener_specs_by_root: Dict[str, list[str]] = {}

        agent_pairs: list[tuple[str, Dict[str, Any]]] = []
        for flow_id, raw in flows_by_namespaced_id.items():
            for node_id, cfg in _collect_agent_nodes(raw):
                agent_pairs.append((flow_id, {"node_id": node_id, "cfg": cfg}))

        if agent_pairs:
            try:
                from abstractagent.adapters.react_runtime import create_react_workflow
                from abstractagent.logic.react import ReActLogic
            except Exception as e:  # pragma: no cover
                raise WorkflowBundleError(
                    "Bundle contains Visual Agent nodes, but AbstractAgent is not installed/importable. "
                    "Install `abstractagent` to execute Agent nodes."
                ) from e

            from abstractcore.tools import ToolDefinition

            try:
                from abstractruntime.integrations.abstractcore.default_tools import list_default_tool_specs
            except Exception as e:  # pragma: no cover
                raise WorkflowBundleError(
                    "Visual Agent nodes require AbstractCore tool schemas (abstractruntime[abstractcore])."
                ) from e

            def _tool_defs_from_specs(specs0: list[dict[str, Any]]) -> list[ToolDefinition]:
                out: list[ToolDefinition] = []
                for s in specs0:
                    if not isinstance(s, dict):
                        continue
                    name = s.get("name")
                    if not isinstance(name, str) or not name.strip():
                        continue
                    desc = s.get("description")
                    params = s.get("parameters")
                    out.append(
                        ToolDefinition(
                            name=name.strip(),
                            description=str(desc or ""),
                            parameters=dict(params) if isinstance(params, dict) else {},
                        )
                    )
                return out

            def _normalize_tool_names(raw_tools: Any) -> list[str]:
                if not isinstance(raw_tools, list):
                    return []
                out: list[str] = []
                for t in raw_tools:
                    if isinstance(t, str) and t.strip():
                        out.append(t.strip())
                return out

            all_tool_defs = _tool_defs_from_specs(list_default_tool_specs())
            # Schema-only builtins (executed as runtime effects by AbstractAgent adapters).
            try:
                from abstractagent.logic.builtins import (  # type: ignore
                    ASK_USER_TOOL,
                    COMPACT_MEMORY_TOOL,
                    DELEGATE_AGENT_TOOL,
                    INSPECT_VARS_TOOL,
                    RECALL_MEMORY_TOOL,
                    REMEMBER_TOOL,
                )

                builtin_defs = [
                    ASK_USER_TOOL,
                    RECALL_MEMORY_TOOL,
                    INSPECT_VARS_TOOL,
                    REMEMBER_TOOL,
                    COMPACT_MEMORY_TOOL,
                    DELEGATE_AGENT_TOOL,
                ]
                seen_names = {t.name for t in all_tool_defs if getattr(t, "name", None)}
                for t in builtin_defs:
                    if getattr(t, "name", None) and t.name not in seen_names:
                        all_tool_defs.append(t)
                        seen_names.add(t.name)
            except Exception:
                pass

            logic = ReActLogic(tools=all_tool_defs)

            from abstractruntime.visualflow_compiler.visual.agent_ids import visual_react_workflow_id

            for flow_id, meta in agent_pairs:
                node_id = str(meta.get("node_id") or "").strip()
                cfg = meta.get("cfg") if isinstance(meta.get("cfg"), dict) else {}
                cfg2 = dict(cfg) if isinstance(cfg, dict) else {}
                workflow_id_raw = cfg2.get("_react_workflow_id")
                react_workflow_id = (
                    workflow_id_raw.strip()
                    if isinstance(workflow_id_raw, str) and workflow_id_raw.strip()
                    else visual_react_workflow_id(flow_id=flow_id, node_id=node_id)
                )
                tools_selected = _normalize_tool_names(cfg2.get("tools"))
                spec = create_react_workflow(
                    logic=logic,
                    workflow_id=react_workflow_id,
                    provider=None,
                    model=None,
                    allowed_tools=tools_selected,
                )
                wf_reg.register(spec)
                specs[str(spec.workflow_id)] = spec

        # Custom event listeners ("On Event" nodes) are compiled into dedicated listener workflows.
        for flow_id, raw in flows_by_namespaced_id.items():
            nodes = raw.get("nodes")
            if not isinstance(nodes, list):
                continue
            for n in nodes:
                if _node_type_from_raw(n) != "on_event":
                    continue
                if not isinstance(n, dict):
                    continue
                node_id = str(n.get("id") or "").strip()
                if not node_id:
                    continue
                listener_wid = _visual_event_listener_workflow_id(flow_id=flow_id, node_id=node_id)

                # Derive a listener workflow with entryNode = on_event node.
                derived: Dict[str, Any] = dict(raw)
                derived["id"] = listener_wid
                derived["entryNode"] = node_id
                try:
                    spec = compile_visualflow(derived)
                except Exception as e:
                    raise WorkflowBundleError(f"Failed compiling On Event listener '{listener_wid}': {e}") from e
                wf_reg.register(spec)
                specs[str(spec.workflow_id)] = spec
                event_listener_specs_by_root.setdefault(flow_id, []).append(str(spec.workflow_id))

        return WorkflowBundleGatewayHost(
            bundles_dir=base,
            data_dir=data_root,
            dynamic_flows_dir=dynamic_dir,
            deprecation_store=dep_store,
            bundles=bundles_by_id,
            latest_bundle_versions=latest_versions,
            runtime=runtime,
            workflow_registry=wf_reg,
            specs=specs,
            event_listener_specs_by_root=event_listener_specs_by_root,
            memory_store=memory_store_obj,
            memory_store_info=memory_store_info,
            _default_bundle_id=default_bundle_id,
        )

    @property
    def run_store(self) -> Any:
        return self.runtime.run_store

    @property
    def ledger_store(self) -> Any:
        return self.runtime.ledger_store

    @property
    def artifact_store(self) -> Any:
        return self.runtime.artifact_store

    def reload_bundles_from_disk(self) -> Dict[str, Any]:
        """Reload bundles/specs from bundles_dir (best-effort, intended for dev).

        Notes:
        - This rebuilds the in-memory registry and swaps host internals in-place so the
          runner can keep using the same host object.
        - Dynamic flows persisted in `data_dir/dynamic_flows` are reloaded as part of
          the rebuild.
        """
        new_host = WorkflowBundleGatewayHost.load_from_dir(
            bundles_dir=self.bundles_dir,
            data_dir=self.data_dir,
            run_store=self.run_store,
            ledger_store=self.ledger_store,
            artifact_store=self.artifact_store,
        )
        with self._lock:
            old_memory_store = getattr(self, "memory_store", None)
            self.bundles = new_host.bundles
            self.latest_bundle_versions = new_host.latest_bundle_versions
            self.runtime = new_host.runtime
            self.workflow_registry = new_host.workflow_registry
            self.specs = new_host.specs
            self.event_listener_specs_by_root = new_host.event_listener_specs_by_root
            self.memory_store = new_host.memory_store
            self.memory_store_info = new_host.memory_store_info
            self._default_bundle_id = new_host._default_bundle_id
            self.deprecation_store = new_host.deprecation_store
        try:
            if old_memory_store is not None and old_memory_store is not getattr(self, "memory_store", None):
                close = getattr(old_memory_store, "close", None)
                if callable(close):
                    close()
        except Exception:
            pass
        bundle_ids = sorted([str(k) for k in (self.bundles or {}).keys() if isinstance(k, str)])
        return {"ok": True, "bundle_ids": bundle_ids, "count": len(bundle_ids)}

    def start_run(
        self,
        *,
        flow_id: str,
        input_data: Dict[str, Any],
        actor_id: str = "gateway",
        bundle_id: Optional[str] = None,
        bundle_version: Optional[str] = None,
        session_id: Optional[str] = None,
    ) -> str:
        fid_raw = str(flow_id or "").strip()

        bid_raw = str(bundle_id or "").strip() if isinstance(bundle_id, str) else ""
        bid_base, bid_ver = _split_bundle_ref(bid_raw)

        bver_raw = str(bundle_version or "").strip() if isinstance(bundle_version, str) else ""
        if bid_ver and bver_raw and bid_ver != bver_raw:
            raise ValueError("bundle_version conflicts with bundle_id (bundle_id already includes '@version')")

        # Effective requested version (may still be None; default to latest per bundle_id).
        requested_ver = bver_raw or bid_ver

        def _get_bundle(*, bundle_id2: str, bundle_version2: Optional[str]) -> tuple[str, WorkflowBundle]:
            bid2 = str(bundle_id2 or "").strip()
            if not bid2:
                raise ValueError("bundle_id is required")
            versions = self.bundles.get(bid2)
            if not isinstance(versions, dict) or not versions:
                raise KeyError(f"Bundle '{bid2}' not found")
            ver2 = str(bundle_version2 or "").strip() if isinstance(bundle_version2, str) and str(bundle_version2).strip() else ""
            if not ver2:
                ver2 = str(self.latest_bundle_versions.get(bid2) or "").strip()
            if not ver2:
                raise KeyError(f"Bundle '{bid2}' has no versions loaded")
            bundle2 = versions.get(ver2)
            if bundle2 is None:
                raise KeyError(f"Bundle '{bid2}@{ver2}' not found")
            return (ver2, bundle2)

        # Default entrypoint selection for the common case:
        # start {bundle_id, input_data} without needing flow_id.
        if not fid_raw:
            selected_bundle_id = bid_base or (str(self._default_bundle_id or "").strip() if self._default_bundle_id else "")
            if not selected_bundle_id:
                raise ValueError(
                    "flow_id is required when multiple bundles are loaded; "
                    "provide bundle_id (or pass flow_id as 'bundle:flow')"
                )
            selected_ver, bundle2 = _get_bundle(bundle_id2=selected_bundle_id, bundle_version2=requested_ver)
            entrypoints = list(getattr(bundle2.manifest, "entrypoints", None) or [])
            default_ep = str(getattr(bundle2.manifest, "default_entrypoint", "") or "").strip()
            if len(entrypoints) == 1:
                ep_fid = str(getattr(entrypoints[0], "flow_id", "") or "").strip()
            elif default_ep:
                ep_fid = default_ep
            else:
                raise ValueError(
                    f"Bundle '{selected_bundle_id}@{selected_ver}' has {len(entrypoints)} entrypoints; "
                    "specify flow_id to select which entrypoint to start "
                    "(or set manifest.default_entrypoint)"
                )
            if not ep_fid:
                raise ValueError(f"Bundle '{selected_bundle_id}@{selected_ver}' entrypoint flow_id is empty")
            workflow_id = _namespace(_bundle_ref(selected_bundle_id, selected_ver), ep_fid)
        else:
            # Allow passing fully qualified flow_id:
            # - bundle:flow (defaults to latest version)
            # - bundle@ver:flow (exact)
            if ":" in fid_raw:
                prefix, inner = fid_raw.split(":", 1)
                prefix_base, prefix_ver = _split_bundle_ref(prefix)
                # Dynamic workflows often use ':' in their ids (e.g. scheduled:uuid).
                # Only treat this as a bundle namespace when the prefix matches a loaded bundle id.
                if prefix_base and prefix_base in self.bundles:
                    if bid_base and prefix_base and bid_base != prefix_base:
                        raise ValueError("flow_id bundle prefix does not match bundle_id")
                    if prefix_ver and requested_ver and prefix_ver != requested_ver:
                        raise ValueError("flow_id version does not match bundle_version")
                    selected_bundle_id = prefix_base or bid_base
                    if not selected_bundle_id:
                        raise ValueError("bundle_id is required")
                    selected_ver, _bundle2 = _get_bundle(
                        bundle_id2=selected_bundle_id,
                        bundle_version2=prefix_ver or requested_ver,
                    )
                    workflow_id = _namespace(_bundle_ref(selected_bundle_id, selected_ver), inner.strip())
                else:
                    if bid_base or requested_ver:
                        raise ValueError(
                            f"flow_id '{fid_raw}' is already namespaced, but bundle_id/bundle_version was also provided"
                        )
                    workflow_id = fid_raw
            else:
                selected_bundle_id = bid_base or (str(self._default_bundle_id or "").strip() if self._default_bundle_id else "")
                if not selected_bundle_id:
                    raise ValueError("bundle_id is required when multiple bundles are loaded (or pass flow_id as 'bundle:flow')")
                selected_ver, _bundle2 = _get_bundle(bundle_id2=selected_bundle_id, bundle_version2=requested_ver)
                workflow_id = _namespace(_bundle_ref(selected_bundle_id, selected_ver), fid_raw)

        # Enforce workflow deprecations (bundle-owned entry workflows).
        # This must live in the host so scheduled child launches are also blocked.
        if ":" in workflow_id:
            prefix, inner = workflow_id.split(":", 1)
            dep_bid, _dep_ver = _split_bundle_ref(prefix)
            dep_flow = inner.strip()
            if dep_bid and dep_flow and dep_bid in self.bundles:
                rec = self.deprecation_store.get_record(bundle_id=dep_bid, flow_id=dep_flow)
                if rec is not None:
                    raise WorkflowDeprecatedError(bundle_id=dep_bid, flow_id=dep_flow, record=rec)

        spec = self.specs.get(workflow_id)
        if spec is None:
            raise KeyError(f"Workflow '{workflow_id}' not found")
        sid = str(session_id).strip() if isinstance(session_id, str) and session_id.strip() else None
        vars0 = dict(input_data or {})
        rt_ns = _ensure_runtime_namespace(vars0)

        # Gateway-owned deployment settings are handed to Runtime explicitly as
        # JSON-safe run state. Lower packages should not read ABSTRACTGATEWAY_*
        # environment names directly.
        prompt_cache_raw = _env("ABSTRACTGATEWAY_PROMPT_CACHE")
        prompt_cache_enabled = _bool_text(prompt_cache_raw)
        if prompt_cache_enabled is not None and not isinstance(rt_ns.get("prompt_cache"), dict):
            rt_ns["prompt_cache"] = {"enabled": bool(prompt_cache_enabled), "version": 1}

        max_attachment_bytes = _int_text(_env("ABSTRACTGATEWAY_MAX_ATTACHMENT_BYTES"))
        if max_attachment_bytes is not None and "max_attachment_bytes" not in rt_ns:
            rt_ns["max_attachment_bytes"] = int(max_attachment_bytes)

        if "workflow_bundles_dir" not in rt_ns:
            rt_ns["workflow_bundles_dir"] = str(self.bundles_dir)

        # Best-effort: seed durable run vars with the gateway runtime defaults so VisualFlow
        # nodes (notably Agent nodes) can inherit provider/model without per-node wiring.
        try:
            cfg = getattr(self.runtime, "config", None)
            default_provider = getattr(cfg, "provider", None)
            default_model = getattr(cfg, "model", None)
        except Exception:  # pragma: no cover
            default_provider = None
            default_model = None
        if default_provider or default_model:
            if isinstance(default_provider, str) and default_provider.strip() and not str(rt_ns.get("provider") or "").strip():
                rt_ns["provider"] = default_provider.strip().lower()
            if isinstance(default_model, str) and default_model.strip() and not str(rt_ns.get("model") or "").strip():
                rt_ns["model"] = default_model.strip()

        run_id = str(self.runtime.start(workflow=spec, vars=vars0, actor_id=actor_id, session_id=sid))

        # Default session_id to the root run_id for durable session-scoped behavior
        # (matches VisualSessionRunner semantics).
        effective_session_id = sid
        if effective_session_id is None:
            try:
                state = self.runtime.get_state(run_id)
                if not getattr(state, "session_id", None):
                    state.session_id = run_id  # type: ignore[attr-defined]
                    self.runtime.run_store.save(state)
                effective_session_id = str(getattr(state, "session_id", None) or run_id).strip() or run_id
            except Exception:
                effective_session_id = run_id

        # Persist a workflow snapshot for reproducible replay (best-effort).
        try:
            if ":" in workflow_id:
                prefix, inner = workflow_id.split(":", 1)
                bid_base, bid_ver2 = _split_bundle_ref(prefix)
                if bid_base and bid_ver2:
                    versions = self.bundles.get(bid_base)
                    bundle2 = versions.get(bid_ver2) if isinstance(versions, dict) else None
                    if bundle2 is not None:
                        bundle_ref = _bundle_ref(bid_base, bid_ver2)
                        man = bundle2.manifest
                        flow_ids = set(man.flows.keys()) if isinstance(getattr(man, "flows", None), dict) else set()
                        id_map = {fid: _namespace(bundle_ref, fid) for fid in flow_ids if isinstance(fid, str) and fid.strip()}
                        rel = man.flow_path_for(inner) if hasattr(man, "flow_path_for") else None
                        raw = bundle2.read_json(rel) if isinstance(rel, str) and rel.strip() else None
                        if isinstance(raw, dict):
                            namespaced_raw = _namespace_visualflow_raw(
                                raw=raw,
                                bundle_id=bundle_ref,
                                flow_id=inner,
                                id_map=id_map,
                            )
                            snapshot = {
                                "kind": "visualflow_json",
                                "bundle_ref": bundle_ref,
                                "flow_id": str(inner),
                                "visualflow": namespaced_raw,
                            }
                            persist_workflow_snapshot(
                                run_store=self.run_store,
                                artifact_store=self.artifact_store,
                                run_id=str(run_id),
                                workflow_id=str(workflow_id),
                                snapshot=snapshot,
                                format="visualflow_json",
                            )
        except Exception:
            pass

        # Start session-scoped event listener workflows (best-effort).
        listener_vars: Dict[str, Any] = {}
        try:
            rt_seed = vars0.get("_runtime")
            if isinstance(rt_seed, dict) and rt_seed:
                listener_vars["_runtime"] = dict(rt_seed)
            limits_seed = vars0.get("_limits")
            if isinstance(limits_seed, dict) and limits_seed:
                listener_vars["_limits"] = dict(limits_seed)
        except Exception:
            listener_vars = {}

        listener_ids = self.event_listener_specs_by_root.get(workflow_id) or []
        for wid in listener_ids:
            listener_spec = self.specs.get(wid)
            if listener_spec is None:
                continue
            try:
                child_run_id = self.runtime.start(
                    workflow=listener_spec,
                    vars=dict(listener_vars),
                    session_id=effective_session_id,
                    parent_run_id=run_id,
                    actor_id=actor_id,
                )
                # Advance to the first WAIT_EVENT.
                self.runtime.tick(workflow=listener_spec, run_id=child_run_id, max_steps=10)
            except Exception:
                continue

        return run_id

    def runtime_and_workflow_for_run(self, run_id: str) -> tuple[Runtime, Any]:
        run = self.run_store.load(str(run_id))
        if run is None:
            raise KeyError(f"Run '{run_id}' not found")
        workflow_id = getattr(run, "workflow_id", None)
        if not isinstance(workflow_id, str) or not workflow_id:
            raise ValueError(f"Run '{run_id}' missing workflow_id")
        spec = self.specs.get(workflow_id)
        if spec is None and ":" in workflow_id:
            # Backward compatibility: older runs may store workflow_id as "bundle:flow"
            # (without bundle_version). Best-effort map to the latest loaded version.
            prefix, fid = workflow_id.split(":", 1)
            prefix_base, prefix_ver = _split_bundle_ref(prefix)
            if prefix_base and not prefix_ver:
                latest = str(self.latest_bundle_versions.get(prefix_base) or "").strip()
                if latest:
                    workflow_id2 = _namespace(_bundle_ref(prefix_base, latest), fid.strip())
                    spec = self.specs.get(workflow_id2)
        if spec is None:
            spec = self._try_register_dynamic_workflow_from_disk(workflow_id)
        if spec is None:
            raise KeyError(f"Workflow '{workflow_id}' not registered")
        return (self.runtime, spec)



from __future__ import annotations

import json
import os
from pathlib import Path

import pytest
from fastapi.testclient import TestClient


@pytest.mark.basic
def test_process_env_endpoints_disabled_by_default(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    runtime_dir = tmp_path / "runtime"
    flows_dir = tmp_path / "flows"
    flows_dir.mkdir(parents=True, exist_ok=True)

    token = "t"
    monkeypatch.setenv("ABSTRACTGATEWAY_DATA_DIR", str(runtime_dir))
    monkeypatch.setenv("ABSTRACTGATEWAY_FLOWS_DIR", str(flows_dir))
    monkeypatch.setenv("ABSTRACTGATEWAY_WORKFLOW_SOURCE", "bundle")
    monkeypatch.setenv("ABSTRACTGATEWAY_AUTH_TOKEN", token)
    monkeypatch.setenv("ABSTRACTGATEWAY_ALLOWED_ORIGINS", "*")
    monkeypatch.setenv("ABSTRACTGATEWAY_RUNNER", "0")
    monkeypatch.setenv("ABSTRACTGATEWAY_ENABLE_PROCESS_MANAGER", "0")

    from abstractgateway.app import app

    headers = {"Authorization": f"Bearer {token}"}
    with TestClient(app) as client:
        r = client.get("/api/gateway/processes/env", headers=headers)
        assert r.status_code == 200, r.text
        body = r.json()
        assert body.get("enabled") is False
        assert body.get("vars") == []

        r2 = client.post("/api/gateway/processes/env", headers=headers, json={"set": {"ABSTRACT_EMAIL_FROM": "x"}})
        # process manager is disabled => 404
        assert r2.status_code == 404, r2.text


@pytest.mark.integration
def test_process_env_endpoints_write_only_and_allowlisted(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    runtime_dir = tmp_path / "runtime"
    flows_dir = tmp_path / "flows"
    flows_dir.mkdir(parents=True, exist_ok=True)

    token = "t"
    monkeypatch.setenv("ABSTRACTGATEWAY_DATA_DIR", str(runtime_dir))
    monkeypatch.setenv("ABSTRACTGATEWAY_FLOWS_DIR", str(flows_dir))
    monkeypatch.setenv("ABSTRACTGATEWAY_WORKFLOW_SOURCE", "bundle")
    monkeypatch.setenv("ABSTRACTGATEWAY_AUTH_TOKEN", token)
    monkeypatch.setenv("ABSTRACTGATEWAY_ALLOWED_ORIGINS", "*")
    monkeypatch.setenv("ABSTRACTGATEWAY_RUNNER", "0")

    monkeypatch.setenv("ABSTRACTGATEWAY_ENABLE_PROCESS_MANAGER", "1")

    from abstractgateway.app import app

    headers = {"Authorization": f"Bearer {token}"}
    with TestClient(app) as client:
        # Initial list includes allowlisted keys.
        r0 = client.get("/api/gateway/processes/env", headers=headers)
        assert r0.status_code == 200, r0.text
        body0 = r0.json()
        assert body0.get("enabled") is True
        keys = {it.get("key") for it in (body0.get("vars") or []) if isinstance(it, dict)}
        assert "ABSTRACT_EMAIL_FROM" in keys
        assert "ABSTRACT_EMAIL_ACCOUNTS_CONFIG" in keys

        # Set a value: response must not contain it.
        secret_value = "secret@example.com"
        config_path = "configs/emails.yaml"
        r1 = client.post(
            "/api/gateway/processes/env",
            headers=headers,
            json={"set": {"ABSTRACT_EMAIL_FROM": secret_value, "ABSTRACT_EMAIL_ACCOUNTS_CONFIG": config_path}},
        )
        assert r1.status_code == 200, r1.text
        body1 = r1.json()
        dumped = json.dumps(body1)
        assert secret_value not in dumped
        assert config_path not in dumped

        # Verify source updated.
        items1 = [it for it in (body1.get("vars") or []) if isinstance(it, dict) and it.get("key") == "ABSTRACT_EMAIL_FROM"]
        assert items1 and items1[0].get("source") == "override"
        items1b = [
            it for it in (body1.get("vars") or []) if isinstance(it, dict) and it.get("key") == "ABSTRACT_EMAIL_ACCOUNTS_CONFIG"
        ]
        assert items1b and items1b[0].get("source") == "override"

        # Persisted on disk (gateway host store).
        path = runtime_dir / "process_manager" / "env_overrides.json"
        assert path.exists()
        obj = json.loads(path.read_text(encoding="utf-8"))
        v = obj.get("vars", {}).get("ABSTRACT_EMAIL_FROM", {})
        assert v.get("enabled") is True
        assert v.get("value") == secret_value
        v_cfg = obj.get("vars", {}).get("ABSTRACT_EMAIL_ACCOUNTS_CONFIG", {})
        assert v_cfg.get("enabled") is True
        assert v_cfg.get("value") == config_path

        # Unset clears stored value.
        r2 = client.post("/api/gateway/processes/env", headers=headers, json={"unset": ["ABSTRACT_EMAIL_FROM"]})
        assert r2.status_code == 200, r2.text
        body2 = r2.json()
        items2 = [it for it in (body2.get("vars") or []) if isinstance(it, dict) and it.get("key") == "ABSTRACT_EMAIL_FROM"]
        assert items2 and items2[0].get("source") == "unset"

        obj2 = json.loads(path.read_text(encoding="utf-8"))
        v2 = obj2.get("vars", {}).get("ABSTRACT_EMAIL_FROM", {})
        assert v2.get("enabled") is False
        assert v2.get("value") == ""

        # Disallowed keys rejected.
        r3 = client.post("/api/gateway/processes/env", headers=headers, json={"set": {"PATH": "/tmp"}})
        assert r3.status_code == 400, r3.text


@pytest.mark.basic
def test_env_overrides_are_applied_on_gateway_startup_when_enabled(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    runtime_dir = tmp_path / "runtime"
    flows_dir = tmp_path / "flows"
    flows_dir.mkdir(parents=True, exist_ok=True)

    # Persist an override before the gateway starts.
    state_dir = runtime_dir / "process_manager"
    state_dir.mkdir(parents=True, exist_ok=True)
    overrides_path = state_dir / "env_overrides.json"
    overrides_path.write_text(
        json.dumps(
            {
                "version": 1,
                "updated_at": "2026-02-06T00:00:00Z",
                "vars": {"ABSTRACT_EMAIL_FROM": {"enabled": True, "value": "persisted@example.com", "updated_at": "2026-02-06T00:00:00Z"}},
            },
            indent=2,
        )
        + "\n",
        encoding="utf-8",
    )

    # Ensure the host environment doesn't already provide it.
    monkeypatch.delenv("ABSTRACT_EMAIL_FROM", raising=False)

    token = "t"
    monkeypatch.setenv("ABSTRACTGATEWAY_DATA_DIR", str(runtime_dir))
    monkeypatch.setenv("ABSTRACTGATEWAY_FLOWS_DIR", str(flows_dir))
    monkeypatch.setenv("ABSTRACTGATEWAY_WORKFLOW_SOURCE", "bundle")
    monkeypatch.setenv("ABSTRACTGATEWAY_AUTH_TOKEN", token)
    monkeypatch.setenv("ABSTRACTGATEWAY_ALLOWED_ORIGINS", "*")
    monkeypatch.setenv("ABSTRACTGATEWAY_RUNNER", "0")
    monkeypatch.setenv("ABSTRACTGATEWAY_ENABLE_PROCESS_MANAGER", "1")

    from abstractgateway.app import app

    headers = {"Authorization": f"Bearer {token}"}
    try:
        with TestClient(app) as client:
            r = client.get("/api/gateway/processes/env", headers=headers)
            assert r.status_code == 200, r.text
            assert os.getenv("ABSTRACT_EMAIL_FROM") == "persisted@example.com"
    finally:
        os.environ.pop("ABSTRACT_EMAIL_FROM", None)



from __future__ import annotations

from pathlib import Path

import pytest


pytestmark = pytest.mark.integration


def test_backlog_exec_runner_persists_logs_to_ledger(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("ABSTRACTGATEWAY_STORE_BACKEND", "file")

    gateway_data_dir = tmp_path / "gateway"
    request_id = "00000000-0000-0000-0000-000000000123"
    run_dir = gateway_data_dir / "backlog_exec_runs" / request_id
    run_dir.mkdir(parents=True, exist_ok=True)

    events_text = '{"type":"test","msg":"EVENTS"}\n'
    stderr_text = "STDERR\n"
    last_msg_text = "LAST_MESSAGE\n"
    (run_dir / "codex_events.jsonl").write_text(events_text, encoding="utf-8")
    (run_dir / "codex_stderr.log").write_text(stderr_text, encoding="utf-8")
    (run_dir / "codex_last_message.txt").write_text(last_msg_text, encoding="utf-8")

    from abstractgateway.maintenance.backlog_exec_runner import _store_backlog_exec_logs_to_ledger  # type: ignore
    from abstractruntime.storage.artifacts import FileArtifactStore  # type: ignore

    req = {
        "request_id": request_id,
        "started_at": "2026-02-02T00:00:00+00:00",
        "finished_at": "2026-02-02T00:00:01+00:00",
        "run_dir_relpath": f"backlog_exec_runs/{request_id}",
        "backlog": {"kind": "planned", "filename": "123-framework-test.md", "relpath": "docs/backlog/planned/123-framework-test.md"},
        "executor": {"type": "codex_cli", "version": "v0"},
    }
    result = {"ok": True, "exit_code": 0, "error": None, "started_at": req["started_at"], "finished_at": req["finished_at"]}

    out = _store_backlog_exec_logs_to_ledger(
        gateway_data_dir=gateway_data_dir,
        request_id=request_id,
        req=req,
        prompt="test prompt",
        run_dir=run_dir,
        result=result,
    )

    ledger = out.get("ledger_run_id")
    assert ledger == f"backlog_exec_{request_id}"

    # Run + ledger files exist (file-backed stores).
    assert (gateway_data_dir / f"run_{ledger}.json").exists()
    assert (gateway_data_dir / f"ledger_{ledger}.jsonl").exists()

    artifacts = out.get("log_artifacts") or {}
    events_ref = artifacts.get("events") or {}
    stderr_ref = artifacts.get("stderr") or {}
    last_ref = artifacts.get("last_message") or {}

    events_id = str(events_ref.get("$artifact") or "").strip()
    stderr_id = str(stderr_ref.get("$artifact") or "").strip()
    last_id = str(last_ref.get("$artifact") or "").strip()
    assert events_id and stderr_id and last_id

    store = FileArtifactStore(gateway_data_dir)
    assert store.load_text(events_id) == events_text
    assert store.load_text(stderr_id) == stderr_text
    assert store.load_text(last_id) == last_msg_text



from __future__ import annotations

import json
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Optional

import pytest
from fastapi import FastAPI
from fastapi.testclient import TestClient

from abstractruntime.core.models import RunStatus


@dataclass
class _Config:
    runner_enabled: bool = True
    data_dir: Path = Path(".")


class _Runner:
    def start(self) -> None:
        return


@dataclass
class _Run:
    status: RunStatus
    output: Dict[str, Any]
    error: str = ""
    waiting: Any = None


class _RunStore:
    def __init__(self) -> None:
        self._runs: Dict[str, _Run] = {}

    def load(self, run_id: str) -> Optional[_Run]:
        return self._runs.get(str(run_id))


class _Host:
    def __init__(self) -> None:
        self.bundles: Dict[str, Any] = {}
        self.run_store = _RunStore()
        self.starts: List[Dict[str, Any]] = []

    def start_run(
        self,
        *,
        flow_id: str,
        bundle_id: Optional[str],
        bundle_version: Optional[str],
        input_data: Dict[str, Any],
        actor_id: str,
        session_id: Optional[str],
    ) -> str:
        del flow_id, bundle_id, bundle_version, actor_id, session_id
        self.starts.append({"input_data": dict(input_data)})
        run_id = f"run-{len(self.starts)}"

        schema = input_data.get("resp_schema") or {}
        required = set(schema.get("required") or [])
        if "draft_markdown" in required:
            resp = {"reply": "ok", "draft_markdown": "# draft\n"}
        else:
            resp = {"reply": "ok"}

        self.run_store._runs[run_id] = _Run(status=RunStatus.COMPLETED, output={"response": json.dumps(resp)})
        return run_id


@dataclass
class _Service:
    config: _Config
    runner: _Runner
    host: _Host
    stores: Any = None


def _make_app(*, monkeypatch: pytest.MonkeyPatch, svc: _Service) -> FastAPI:
    import abstractgateway.routes.gateway as gateway_routes

    monkeypatch.setattr(gateway_routes, "get_gateway_service", lambda: svc)

    app = FastAPI()
    app.include_router(gateway_routes.router, prefix="/api")
    return app


@pytest.mark.basic
def test_backlog_advisor_readonly_tools_and_allowed_paths(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    repo_root = tmp_path / "repo"
    (repo_root / "docs" / "backlog").mkdir(parents=True, exist_ok=True)
    monkeypatch.setenv("ABSTRACTGATEWAY_TRIAGE_REPO_ROOT", str(repo_root))
    monkeypatch.setenv("ABSTRACTGATEWAY_PROVIDER", "stub")
    monkeypatch.setenv("ABSTRACTGATEWAY_MODEL", "stub-model")

    gateway_dir = tmp_path / "gateway_data"
    gateway_dir.mkdir(parents=True, exist_ok=True)

    host = _Host()
    svc = _Service(config=_Config(data_dir=gateway_dir), runner=_Runner(), host=host)
    app = _make_app(monkeypatch=monkeypatch, svc=svc)

    with TestClient(app) as client:
        r = client.post("/api/gateway/backlog/advisor", json={"message": "hi", "messages": [{"role": "user", "content": "hi"}]})
        assert r.status_code == 200
        assert r.json().get("ok") is True

    assert host.starts, "expected advisor to start a run"
    input_data = host.starts[-1]["input_data"]

    # Read-only tool allowlist should include repo inspection tools (and exclude writes/exec).
    tools = list(input_data.get("tools") or [])
    assert "read_file" in tools
    assert "search_files" in tools
    assert "list_files" in tools
    assert "skim_folders" in tools
    assert "skim_files" in tools
    assert "analyze_code" in tools
    assert "write_file" not in tools
    assert "edit_file" not in tools
    assert "execute_command" not in tools

    # Gateway data dir should be available via allowed paths even if outside the repo root.
    assert input_data.get("workspace_access_mode") == "workspace_or_allowed"
    allowed = list(input_data.get("workspace_allowed_paths") or [])
    assert str(gateway_dir.resolve()) in allowed

    # Advisor prompt context should mention gateway_data_dir.
    prompt = str(input_data.get("prompt") or "")
    assert "gateway_data_dir" in prompt


@pytest.mark.basic
def test_backlog_maintain_readonly_tools_and_allowed_paths(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    repo_root = tmp_path / "repo"
    backlog_root = repo_root / "docs" / "backlog"
    (backlog_root / "proposed").mkdir(parents=True, exist_ok=True)
    (backlog_root / "template.md").write_text("# {ID}-{Package}: {Title}\n", encoding="utf-8")
    (backlog_root / "proposed" / "672-framework-test.md").write_text("# 672-framework: [TASK] Test\n", encoding="utf-8")
    monkeypatch.setenv("ABSTRACTGATEWAY_TRIAGE_REPO_ROOT", str(repo_root))
    monkeypatch.setenv("ABSTRACTGATEWAY_PROVIDER", "stub")
    monkeypatch.setenv("ABSTRACTGATEWAY_MODEL", "stub-model")

    gateway_dir = tmp_path / "gateway_data"
    gateway_dir.mkdir(parents=True, exist_ok=True)

    host = _Host()
    svc = _Service(config=_Config(data_dir=gateway_dir), runner=_Runner(), host=host)
    app = _make_app(monkeypatch=monkeypatch, svc=svc)

    with TestClient(app) as client:
        r = client.post(
            "/api/gateway/backlog/maintain",
            json={
                "kind": "proposed",
                "filename": "672-framework-test.md",
                "package": "framework",
                "title": "Test",
                "summary": "Test",
                "draft_markdown": "# 672-framework: [TASK] Test\n",
                "messages": [{"role": "user", "content": "please improve"}],
            },
        )
        assert r.status_code == 200
        assert r.json().get("ok") is True

    assert host.starts, "expected maintainer to start a run"
    input_data = host.starts[-1]["input_data"]

    tools = list(input_data.get("tools") or [])
    assert "read_file" in tools
    assert "search_files" in tools
    assert "list_files" in tools
    assert "skim_folders" in tools
    assert "skim_files" in tools
    assert "analyze_code" in tools
    assert "web_search" in tools
    assert "fetch_url" in tools
    assert "write_file" not in tools
    assert "edit_file" not in tools
    assert "execute_command" not in tools

    assert input_data.get("workspace_access_mode") == "workspace_or_allowed"
    allowed = list(input_data.get("workspace_allowed_paths") or [])
    assert str(gateway_dir.resolve()) in allowed



"""
Tests for gateway VisualFlow helper functions.
"""

from __future__ import annotations

import sys
from pathlib import Path

import pytest
from fastapi import HTTPException


ROOT = Path(__file__).resolve().parents[2]
sys.path.append(str(ROOT / "abstractgateway" / "src"))

from abstractgateway.routes import gateway as gw  # noqa: E402


def test_safe_visualflow_id_accepts_safe_values() -> None:
    assert gw._safe_visualflow_id("flow-1_ok") == "flow-1_ok"
    assert gw._safe_visualflow_id("flow.2") == "flow.2"


def test_safe_visualflow_id_rejects_unsafe_values() -> None:
    with pytest.raises(HTTPException):
        gw._safe_visualflow_id("")
    with pytest.raises(HTTPException):
        gw._safe_visualflow_id("..")
    with pytest.raises(HTTPException):
        gw._safe_visualflow_id("../escape")
    with pytest.raises(HTTPException):
        gw._safe_visualflow_id("bad/segment")


def test_semver_parsing_and_bump_patch() -> None:
    assert gw._try_parse_semver("1.2.3") == (1, 2, 3)
    assert gw._try_parse_semver("1.2") == (1, 2, 0)
    assert gw._try_parse_semver("1") == (1, 0, 0)
    assert gw._try_parse_semver("1.a") is None

    assert gw._bump_patch("1.2.3") == "1.2.4"
    assert gw._bump_patch("1") == "1.0.1"
    assert gw._bump_patch("beta") == "beta.1"
    assert gw._bump_patch("") == "0.0.1"


def test_version_ordering_helpers() -> None:
    versions = [{"bundle_version": "0.1.0"}, {"bundle_version": "0.2.0"}, {"bundle_version": "0.1.5"}]
    assert gw._latest_version(versions) == "0.2.0"

    dated = [
        {"bundle_version": "alpha", "created_at": "2024-01-01T00:00:00Z"},
        {"bundle_version": "beta", "created_at": "2024-02-01T00:00:00Z"},
    ]
    assert gw._latest_version(dated) == "beta"
    assert gw._origin_version(dated) == "alpha"



from __future__ import annotations

import os
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, Optional

from .config import GatewayHostConfig
from .embeddings_config import resolve_embedding_config
from .hosts.visualflow_host import VisualFlowGatewayHost, VisualFlowRegistry
from .runner import GatewayRunner, GatewayRunnerConfig
from .security import GatewayAuthPolicy, load_gateway_auth_policy_from_env
from .stores import GatewayStores, build_file_stores, build_sqlite_stores


@dataclass(frozen=True)
class GatewayService:
    """Composition root: host + runner + security policy."""

    config: GatewayHostConfig
    stores: GatewayStores
    host: Any
    runner: GatewayRunner
    auth_policy: GatewayAuthPolicy
    embedding_provider: Optional[str] = None
    embedding_model: Optional[str] = None
    embeddings_client: Optional[Any] = None
    telegram_bridge: Optional[Any] = None
    email_bridge: Optional[Any] = None


_service: Optional[GatewayService] = None
_backlog_exec_runner: Optional[Any] = None
_backlog_exec_runner_error: Optional[str] = None


def backlog_exec_runner_status() -> Dict[str, Any]:
    runner = _backlog_exec_runner
    alive = False
    last_error: Optional[str] = None
    try:
        if runner is not None and hasattr(runner, "is_running"):
            alive = bool(runner.is_running())
        elif runner is not None and hasattr(runner, "_thread"):
            t = getattr(runner, "_thread", None)
            alive = bool(t is not None and getattr(t, "is_alive", lambda: False)())
    except Exception:
        alive = False

    try:
        if runner is not None and hasattr(runner, "last_error"):
            last_error = runner.last_error()
    except Exception:
        last_error = None

    if _backlog_exec_runner_error:
        last_error = _backlog_exec_runner_error

    return {"alive": bool(alive), "error": (str(last_error) if last_error else "") or None}


def get_gateway_service() -> GatewayService:
    global _service
    if _service is None:
        _service = create_default_gateway_service()
    return _service


def create_default_gateway_service() -> GatewayService:
    cfg = GatewayHostConfig.from_env()
    backend = str(getattr(cfg, "store_backend", "file") or "file").strip().lower() or "file"
    if backend == "file":
        stores = build_file_stores(base_dir=cfg.data_dir)
    elif backend == "sqlite":
        stores = build_sqlite_stores(base_dir=cfg.data_dir, db_path=getattr(cfg, "db_path", None))
    else:
        raise RuntimeError(f"Unsupported store backend: {backend}. Supported: file|sqlite")

    # Best-effort: apply persisted process-manager env overrides early so runtime
    # integrations (email bridge, report triage, etc.) observe the configured values
    # immediately after a gateway restart.
    try:
        enabled_raw = os.getenv("ABSTRACTGATEWAY_ENABLE_PROCESS_MANAGER")
        if enabled_raw is not None and str(enabled_raw).strip().lower() in {"1", "true", "yes", "on"}:
            from .maintenance.process_manager import get_managed_env_var_manager

            get_managed_env_var_manager(base_dir=stores.base_dir)
    except Exception:
        pass

    # Workflow source:
    # - bundle (default): `.flow` bundles with VisualFlow JSON (compiled via AbstractRuntime; no AbstractFlow import)
    # - visualflow (optional): load VisualFlow JSON files directly from a directory (host wiring currently uses AbstractFlow extras)
    source = str(os.getenv("ABSTRACTGATEWAY_WORKFLOW_SOURCE", "bundle") or "bundle").strip().lower()
    if source == "bundle":
        from .hosts.bundle_host import WorkflowBundleGatewayHost

        host = WorkflowBundleGatewayHost.load_from_dir(
            bundles_dir=cfg.flows_dir,
            data_dir=cfg.data_dir,
            run_store=stores.run_store,
            ledger_store=stores.ledger_store,
            artifact_store=stores.artifact_store,
        )
    elif source == "visualflow":
        flows = VisualFlowRegistry(flows_dir=cfg.flows_dir).load()
        host = VisualFlowGatewayHost(
            flows_dir=cfg.flows_dir,
            flows=flows,
            run_store=stores.run_store,
            ledger_store=stores.ledger_store,
            artifact_store=stores.artifact_store,
        )
    else:
        raise RuntimeError(f"Unsupported workflow source: {source}. Supported: bundle|visualflow")

    runner_cfg = GatewayRunnerConfig(
        poll_interval_s=float(cfg.poll_interval_s),
        command_batch_limit=int(cfg.command_batch_limit),
        tick_max_steps=int(cfg.tick_max_steps),
        tick_workers=int(cfg.tick_workers),
        run_scan_limit=int(cfg.run_scan_limit),
    )
    runner = GatewayRunner(
        base_dir=stores.base_dir,
        host=host,
        config=runner_cfg,
        enable=bool(cfg.runner_enabled),
        command_store=stores.command_store,
        cursor_store=stores.command_cursor_store,
    )

    policy = load_gateway_auth_policy_from_env()

    embedding_provider: Optional[str] = None
    embedding_model: Optional[str] = None
    embeddings_client: Optional[Any] = None
    try:
        embedding_provider, embedding_model = resolve_embedding_config(base_dir=stores.base_dir)
        from abstractruntime.integrations.abstractcore.embeddings_client import AbstractCoreEmbeddingsClient

        embeddings_client = AbstractCoreEmbeddingsClient(
            provider=embedding_provider,
            model=embedding_model,
            manager_kwargs={
                "cache_dir": Path(stores.base_dir) / "abstractcore" / "embeddings",
                # Embeddings must be trustworthy for semantic retrieval; do not return zero vectors on failure.
                "strict": True,
            },
        )
    except Exception:
        # Embeddings are optional: the gateway may run without AbstractCore embedding deps.
        embeddings_client = None

    telegram_bridge = None
    enabled_raw = os.getenv("ABSTRACT_TELEGRAM_BRIDGE")
    if enabled_raw is not None and str(enabled_raw).strip().lower() in {"1", "true", "yes", "on"}:
        try:
            from .integrations.telegram_bridge import TelegramBridge, TelegramBridgeConfig
        except Exception as e:
            raise RuntimeError(
                "Telegram bridge is enabled (ABSTRACT_TELEGRAM_BRIDGE=1) but Gateway's Telegram integration could not be imported. "
                "Install/repair the base Gateway environment with: `pip install abstractgateway`"
            ) from e

        tcfg = TelegramBridgeConfig.from_env(base_dir=cfg.data_dir)
        if not tcfg.flow_id:
            raise RuntimeError("ABSTRACT_TELEGRAM_FLOW_ID is required when ABSTRACT_TELEGRAM_BRIDGE=1")
        telegram_bridge = TelegramBridge(config=tcfg, host=host, runner=runner, artifact_store=stores.artifact_store)

    email_bridge = None
    email_enabled_raw = os.getenv("ABSTRACT_EMAIL_BRIDGE")
    if email_enabled_raw is not None and str(email_enabled_raw).strip().lower() in {"1", "true", "yes", "on"}:
        from .integrations.email_bridge import EmailBridge, EmailBridgeConfig

        ecfg = EmailBridgeConfig.from_env(base_dir=cfg.data_dir)
        email_bridge = EmailBridge(config=ecfg, host=host, runner=runner, artifact_store=stores.artifact_store)

    return GatewayService(
        config=cfg,
        stores=stores,
        host=host,
        runner=runner,
        auth_policy=policy,
        embedding_provider=embedding_provider,
        embedding_model=embedding_model,
        embeddings_client=embeddings_client,
        telegram_bridge=telegram_bridge,
        email_bridge=email_bridge,
    )


def start_gateway_runner() -> None:
    svc = get_gateway_service()
    svc.runner.start()
    # Optional: backlog execution runner (consumes backlog_exec_queue and executes requests).
    global _backlog_exec_runner, _backlog_exec_runner_error
    try:
        if _backlog_exec_runner is None:
            from .maintenance.backlog_exec_runner import BacklogExecRunner, BacklogExecRunnerConfig

            cfg = BacklogExecRunnerConfig.from_env()
            if cfg.enabled:
                _backlog_exec_runner = BacklogExecRunner(gateway_data_dir=svc.stores.base_dir, cfg=cfg)
                _backlog_exec_runner.start()
                _backlog_exec_runner_error = None
    except Exception as e:
        # Best-effort: never break the gateway runner start if maintenance runner fails.
        _backlog_exec_runner = None
        try:
            _backlog_exec_runner_error = str(e)
        except Exception:
            pass
    bridge = getattr(svc, "telegram_bridge", None)
    if bridge is not None:
        bridge.start()
    email_bridge = getattr(svc, "email_bridge", None)
    if email_bridge is not None:
        email_bridge.start()


def stop_gateway_runner() -> None:
    global _service, _backlog_exec_runner, _backlog_exec_runner_error
    if _service is None:
        return
    try:
        try:
            bridge = getattr(_service, "telegram_bridge", None)
            if bridge is not None:
                bridge.stop()
        except Exception:
            pass
        try:
            bridge2 = getattr(_service, "email_bridge", None)
            if bridge2 is not None:
                bridge2.stop()
        except Exception:
            pass
        try:
            if _backlog_exec_runner is not None:
                _backlog_exec_runner.stop()
        except Exception:
            pass
        _backlog_exec_runner = None
        _backlog_exec_runner_error = None
        _service.runner.stop()
    finally:
        _service = None


def run_summary(run: Any) -> Dict[str, Any]:
    """HTTP-safe run summary (do not return full run.vars)."""

    waiting = getattr(run, "waiting", None)
    status = getattr(getattr(run, "status", None), "value", None) or str(getattr(run, "status", "unknown"))
    out: Dict[str, Any] = {
        "run_id": getattr(run, "run_id", ""),
        "workflow_id": getattr(run, "workflow_id", None),
        "status": status,
        "current_node": getattr(run, "current_node", None),
        "created_at": getattr(run, "created_at", None),
        "updated_at": getattr(run, "updated_at", None),
        "actor_id": getattr(run, "actor_id", None),
        "session_id": getattr(run, "session_id", None),
        "parent_run_id": getattr(run, "parent_run_id", None),
        "error": getattr(run, "error", None),
        "flow_warnings": None,
        # Best-effort pause metadata. We intentionally do not return full run.vars over HTTP.
        "paused": False,
        "pause_reason": None,
        "paused_at": None,
        "resumed_at": None,
        "waiting": None,
        # Best-effort schedule metadata (only for scheduled parent runs).
        "is_scheduled": False,
        "schedule": None,
        # Best-effort limits metadata (for UX, not for enforcing).
        "limits": None,
    }
    try:
        vars_obj = getattr(run, "vars", None)
        runtime_ns = vars_obj.get("_runtime") if isinstance(vars_obj, dict) else None
        control = runtime_ns.get("control") if isinstance(runtime_ns, dict) else None
        if isinstance(control, dict):
            out["paused"] = bool(control.get("paused") is True)
            out["pause_reason"] = control.get("pause_reason")
            out["paused_at"] = control.get("paused_at")
            out["resumed_at"] = control.get("resumed_at")
    except Exception:
        pass

    try:
        vars_obj = getattr(run, "vars", None)
        raw_warnings = vars_obj.get("_flow_warnings") if isinstance(vars_obj, dict) else None
        if isinstance(raw_warnings, list):
            cleaned: list[str] = []
            for w in raw_warnings:
                if not isinstance(w, str):
                    continue
                s = w.strip()
                if s:
                    cleaned.append(s)
            if cleaned:
                out["flow_warnings"] = cleaned
    except Exception:
        pass

    # Schedule + limits are safe, small subsets for UI. Never return full run.vars.
    try:
        vars_obj = getattr(run, "vars", None)
        if isinstance(vars_obj, dict):
            meta = vars_obj.get("_meta")
            schedule = meta.get("schedule") if isinstance(meta, dict) else None
            if isinstance(schedule, dict) and schedule.get("kind") == "scheduled_run":
                out["is_scheduled"] = True
                out["schedule"] = {
                    "kind": "scheduled_run",
                    "interval": schedule.get("interval"),
                    "repeat_count": schedule.get("repeat_count"),
                    "repeat_until": schedule.get("repeat_until"),
                    "start_at": schedule.get("start_at"),
                    "share_context": schedule.get("share_context"),
                    "target_workflow_id": schedule.get("target_workflow_id"),
                    "target_bundle_ref": schedule.get("target_bundle_ref"),
                    "target_flow_id": schedule.get("target_flow_id"),
                    "created_at": schedule.get("created_at"),
                    "updated_at": schedule.get("updated_at"),
                }
            else:
                wid = getattr(run, "workflow_id", None)
                if isinstance(wid, str) and wid.startswith("scheduled:"):
                    out["is_scheduled"] = True

            limits = vars_obj.get("_limits")
            if isinstance(limits, dict):
                used = limits.get("estimated_tokens_used")
                max_tokens = limits.get("max_tokens")
                max_input = limits.get("max_input_tokens")
                warn_pct = limits.get("warn_tokens_pct")
                budget = max_input if max_input is not None else max_tokens
                pct = None
                try:
                    used_i = int(used) if used is not None and not isinstance(used, bool) else None
                    budget_i = int(budget) if budget is not None and not isinstance(budget, bool) else None
                    if used_i is not None and budget_i is not None and budget_i > 0:
                        pct = float(used_i) / float(budget_i)
                except Exception:
                    pct = None
                out["limits"] = {
                    "tokens": {
                        "estimated_used": used,
                        "max_tokens": max_tokens,
                        "max_input_tokens": max_input,
                        "pct": pct,
                        "warn_tokens_pct": warn_pct,
                    }
                }
    except Exception:
        pass
    if waiting is not None:
        out["waiting"] = {
            "reason": getattr(getattr(waiting, "reason", None), "value", None) or str(getattr(waiting, "reason", "")),
            "wait_key": getattr(waiting, "wait_key", None),
            "until": getattr(waiting, "until", None),
            "prompt": getattr(waiting, "prompt", None),
            "choices": getattr(waiting, "choices", None),
            "allow_free_text": getattr(waiting, "allow_free_text", None),
            "details": getattr(waiting, "details", None),
        }
    return out



from __future__ import annotations

import logging
import os
import subprocess
import sys
import types

import pytest


@pytest.mark.basic
def test_cli_serve_no_runner_sets_env_and_invokes_uvicorn(monkeypatch: pytest.MonkeyPatch) -> None:
    from abstractgateway import cli as gateway_cli

    called: dict[str, object] = {}

    uvicorn = types.ModuleType("uvicorn")

    def _run(app: str, **kwargs: object) -> None:
        called["app"] = app
        called.update(kwargs)
        assert os.environ.get("ABSTRACTGATEWAY_RUNNER") == "0"

    uvicorn.run = _run  # type: ignore[attr-defined]
    monkeypatch.setitem(sys.modules, "uvicorn", uvicorn)

    monkeypatch.setattr(gateway_cli, "_resolve_default_console_level", lambda: logging.ERROR)
    monkeypatch.delenv("ABSTRACTGATEWAY_RUNNER", raising=False)
    monkeypatch.setenv("ABSTRACTGATEWAY_AUTH_TOKEN", "t")
    gateway_cli.main(["serve", "--no-runner", "--host", "127.0.0.1", "--port", "9999"])

    assert called["app"] == "abstractgateway.app:app"
    assert called["host"] == "127.0.0.1"
    assert called["port"] == 9999
    assert called["reload"] is False
    assert called["log_level"] == "error"


@pytest.mark.basic
def test_cli_runner_forces_runner_env_and_calls_start_stop(monkeypatch: pytest.MonkeyPatch) -> None:
    from abstractgateway import cli as gateway_cli
    from abstractgateway import service as gateway_service

    calls: list[str] = []

    def _start() -> None:
        assert os.environ.get("ABSTRACTGATEWAY_RUNNER") == "1"
        calls.append("start")

    def _stop() -> None:
        assert os.environ.get("ABSTRACTGATEWAY_RUNNER") == "1"
        calls.append("stop")

    monkeypatch.setattr(gateway_service, "start_gateway_runner", _start)
    monkeypatch.setattr(gateway_service, "stop_gateway_runner", _stop)

    class _ImmediateEvent:
        def __init__(self) -> None:
            self._set = True

        def is_set(self) -> bool:
            return self._set

        def set(self) -> None:
            self._set = True

        def wait(self, timeout: float | None = None) -> bool:  # noqa: ARG002
            self._set = True
            return True

    monkeypatch.setattr(gateway_cli.threading, "Event", _ImmediateEvent)

    monkeypatch.delenv("ABSTRACTGATEWAY_RUNNER", raising=False)
    gateway_cli.main(["runner"])

    # CLI sets the env var for the process while running, but restores it on exit.
    assert os.environ.get("ABSTRACTGATEWAY_RUNNER") is None
    assert calls == ["start", "stop"]


@pytest.mark.basic
def test_runner_module_does_not_import_fastapi() -> None:
    proc = subprocess.run(
        [sys.executable, "-c", "import sys; import abstractgateway.runner; print('fastapi' in sys.modules)"],
        check=True,
        capture_output=True,
        text=True,
        env=os.environ.copy(),
        timeout=15,
    )
    assert proc.stdout.strip() == "False"


@pytest.mark.basic
def test_cli_serve_requires_auth_token(monkeypatch: pytest.MonkeyPatch) -> None:
    from abstractgateway import cli as gateway_cli

    monkeypatch.delenv("ABSTRACTGATEWAY_AUTH_TOKEN", raising=False)
    with pytest.raises(SystemExit) as e:
        gateway_cli.main(["serve", "--host", "127.0.0.1", "--port", "9999", "--no-runner"])
    assert "Missing gateway auth token" in str(e.value)


@pytest.mark.basic
def test_cli_serve_refuses_weak_token_on_public_bind(monkeypatch: pytest.MonkeyPatch) -> None:
    from abstractgateway import cli as gateway_cli

    # Weak token is allowed on loopback, but should be rejected when binding publicly.
    monkeypatch.setenv("ABSTRACTGATEWAY_AUTH_TOKEN", "t")
    with pytest.raises(SystemExit) as e:
        gateway_cli.main(["serve", "--host", "0.0.0.0", "--port", "9999", "--no-runner"])
    assert "Refusing to start" in str(e.value)



from __future__ import annotations

import datetime
import hashlib
import json
from dataclasses import asdict
from pathlib import Path
from typing import Any, Dict, List, Optional

from .report_models import SimilarityCandidate, TriageDecision


def _now_utc_iso() -> str:
    return datetime.datetime.now(datetime.timezone.utc).isoformat()


def decision_id_for_report(*, report_relpath: str) -> str:
    # Stable id so repeated triage runs update the same decision entry.
    h = hashlib.sha1(str(report_relpath).encode("utf-8")).hexdigest()
    return h[:16]


def decisions_dir(*, gateway_data_dir: Path) -> Path:
    out = Path(gateway_data_dir).expanduser().resolve() / "triage_queue"
    out.mkdir(parents=True, exist_ok=True)
    return out


def _decision_path(*, dir_path: Path, decision_id: str) -> Path:
    safe = "".join([c for c in str(decision_id or "") if c.isalnum() or c in {"_", "-"}]).strip()
    safe = safe or "decision"
    return Path(dir_path) / f"{safe}.json"


def load_decision(*, dir_path: Path, decision_id: str) -> Optional[TriageDecision]:
    path = _decision_path(dir_path=dir_path, decision_id=decision_id)
    if not path.exists():
        return None
    try:
        data = json.loads(path.read_text(encoding="utf-8"))
    except Exception:
        return None
    if not isinstance(data, dict):
        return None

    try:
        dup_raw = data.get("duplicates") or []
        dups: List[SimilarityCandidate] = []
        if isinstance(dup_raw, list):
            for item in dup_raw:
                if not isinstance(item, dict):
                    continue
                kind = str(item.get("kind") or "").strip() or "report"
                ref = str(item.get("ref") or "").strip()
                title = str(item.get("title") or "").strip()
                try:
                    score = float(item.get("score"))
                except Exception:
                    score = 0.0
                if not ref:
                    continue
                dups.append(SimilarityCandidate(kind=kind, ref=ref, score=score, title=title))
    except Exception:
        dups = []

    missing = data.get("missing_fields") or []
    missing2 = [str(m).strip() for m in missing if isinstance(m, str) and m.strip()] if isinstance(missing, list) else []

    decision = TriageDecision(
        decision_id=str(data.get("decision_id") or decision_id),
        report_type=str(data.get("report_type") or "bug"),  # type: ignore[arg-type]
        report_relpath=str(data.get("report_relpath") or ""),
        status=str(data.get("status") or "pending"),  # type: ignore[arg-type]
        created_at=str(data.get("created_at") or ""),
        updated_at=str(data.get("updated_at") or ""),
        defer_until=str(data.get("defer_until") or ""),
        missing_fields=missing2,
        duplicates=dups,
        draft_relpath=str(data.get("draft_relpath") or ""),
        llm_suggestion=data.get("llm_suggestion") if isinstance(data.get("llm_suggestion"), dict) else {},
    )
    return decision


def save_decision(*, dir_path: Path, decision: TriageDecision) -> Path:
    path = _decision_path(dir_path=dir_path, decision_id=decision.decision_id)
    payload: Dict[str, Any] = asdict(decision)
    # Convert dataclass list to dict list.
    payload["duplicates"] = [asdict(d) for d in decision.duplicates]
    if not payload.get("updated_at"):
        payload["updated_at"] = _now_utc_iso()
    if not payload.get("created_at"):
        payload["created_at"] = payload["updated_at"]
    path.write_text(json.dumps(payload, ensure_ascii=False, indent=2, sort_keys=True) + "\n", encoding="utf-8")
    return path


def upsert_decision(
    *,
    dir_path: Path,
    decision_id: str,
    report_type: str,
    report_relpath: str,
    missing_fields: List[str],
    duplicates: List[SimilarityCandidate],
    llm_suggestion: Optional[Dict[str, Any]] = None,
) -> TriageDecision:
    existing = load_decision(dir_path=dir_path, decision_id=decision_id)
    now = _now_utc_iso()

    if existing is None:
        existing = TriageDecision(
            decision_id=decision_id,
            report_type=report_type,  # type: ignore[arg-type]
            report_relpath=report_relpath,
            status="pending",
            created_at=now,
            updated_at=now,
        )

    # Keep status/defer_until/draft_relpath if already set; refresh computed fields.
    existing.missing_fields = list(missing_fields)
    existing.duplicates = list(duplicates)
    existing.updated_at = now
    if llm_suggestion is not None:
        existing.llm_suggestion = dict(llm_suggestion)

    save_decision(dir_path=dir_path, decision=existing)
    return existing


def iter_decisions(dir_path: Path) -> List[TriageDecision]:
    out: List[TriageDecision] = []
    if not dir_path.exists():
        return out
    for p in sorted(dir_path.glob("*.json")):
        did = p.stem
        d = load_decision(dir_path=dir_path, decision_id=did)
        if d is not None:
            out.append(d)
    return out




from __future__ import annotations

import json
from pathlib import Path
import zipfile

import pytest
from fastapi.testclient import TestClient


def _make_min_flow(flow_id: str) -> dict:
    fid = str(flow_id or "").strip() or "root"
    return {
        "id": fid,
        "name": fid,
        "description": "",
        "interfaces": [],
        "nodes": [
            {
                "id": "start",
                "type": "on_flow_start",
                "position": {"x": 0.0, "y": 0.0},
                "data": {"nodeType": "on_flow_start", "label": "On Flow Start", "inputs": [], "outputs": [{"id": "exec-out", "label": "", "type": "execution"}]},
            },
            {
                "id": "end",
                "type": "on_flow_end",
                "position": {"x": 10.0, "y": 0.0},
                "data": {"nodeType": "on_flow_end", "label": "On Flow End", "inputs": [{"id": "exec-in", "label": "", "type": "execution"}], "outputs": []},
            },
        ],
        "edges": [
            {"id": "e1", "source": "start", "sourceHandle": "exec-out", "target": "end", "targetHandle": "exec-in", "animated": True},
        ],
        "entryNode": "start",
    }


def _write_bundle(path: Path, *, bundle_id: str, bundle_version: str, flow_id: str) -> None:
    manifest = {
        "bundle_format_version": "1",
        "bundle_id": bundle_id,
        "bundle_version": bundle_version,
        "created_at": "2026-01-24T00:00:00Z",
        "entrypoints": [{"flow_id": flow_id, "name": flow_id, "description": "", "interfaces": []}],
        "default_entrypoint": flow_id,
        "flows": {flow_id: f"flows/{flow_id}.json"},
        "metadata": {"test": True},
    }
    flow = _make_min_flow(flow_id)
    path.parent.mkdir(parents=True, exist_ok=True)
    with zipfile.ZipFile(path, "w", compression=zipfile.ZIP_DEFLATED) as zf:
        zf.writestr("manifest.json", json.dumps(manifest, ensure_ascii=False, indent=2))
        zf.writestr(f"flows/{flow_id}.json", json.dumps(flow, ensure_ascii=False, indent=2))


@pytest.mark.integration
def test_gateway_bundle_upload_and_remove(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    runtime_dir = tmp_path / "runtime"
    bundles_dir = tmp_path / "bundles"
    bundles_dir.mkdir(parents=True, exist_ok=True)

    token = "t"
    monkeypatch.setenv("ABSTRACTGATEWAY_DATA_DIR", str(runtime_dir))
    monkeypatch.setenv("ABSTRACTGATEWAY_FLOWS_DIR", str(bundles_dir))
    monkeypatch.setenv("ABSTRACTGATEWAY_AUTH_TOKEN", token)
    monkeypatch.setenv("ABSTRACTGATEWAY_WORKFLOW_SOURCE", "bundle")

    bundle_path = tmp_path / "demo@0.0.1.flow"
    _write_bundle(bundle_path, bundle_id="demo", bundle_version="0.0.1", flow_id="root")

    from abstractgateway.app import app

    headers = {"Authorization": f"Bearer {token}"}
    with TestClient(app) as client:
        r0 = client.get("/api/gateway/bundles", headers=headers)
        assert r0.status_code == 200
        assert isinstance(r0.json().get("items"), list)

        up = client.post(
            "/api/gateway/bundles/upload",
            headers=headers,
            data={"overwrite": "false", "reload": "true"},
            files={"file": (bundle_path.name, bundle_path.read_bytes(), "application/octet-stream")},
        )
        assert up.status_code == 200, up.text
        body = up.json()
        assert body.get("ok") is True
        assert body.get("bundle_ref") == "demo@0.0.1"

        r1 = client.get("/api/gateway/bundles?all_versions=true", headers=headers)
        assert r1.status_code == 200
        items = r1.json().get("items") or []
        assert any(it.get("bundle_ref") == "demo@0.0.1" for it in items)

        rm = client.delete("/api/gateway/bundles/demo?bundle_version=0.0.1&reload=true", headers=headers)
        assert rm.status_code == 200, rm.text
        assert rm.json().get("removed") == 1

        r2 = client.get("/api/gateway/bundles?all_versions=true", headers=headers)
        assert r2.status_code == 200
        items2 = r2.json().get("items") or []
        assert not any(it.get("bundle_ref") == "demo@0.0.1" for it in items2)

        rm2 = client.delete("/api/gateway/bundles/demo?bundle_version=0.0.1&reload=true", headers=headers)
        assert rm2.status_code == 404



from __future__ import annotations

from pathlib import Path
from types import SimpleNamespace

import pytest


@pytest.mark.basic
def test_pid_commandline_uses_wide_ps(monkeypatch: pytest.MonkeyPatch) -> None:
    import abstractgateway.maintenance.process_manager as pm

    calls: list[list[str]] = []

    def _fake_run(args, **kwargs):
        calls.append(list(args))
        return SimpleNamespace(stdout="python -c pass\n", returncode=0)

    monkeypatch.setattr(pm.subprocess, "run", _fake_run)

    out = pm._pid_commandline(12345)
    assert out == "python -c pass"
    assert calls
    assert calls[0][0] == "ps"
    assert "-ww" in calls[0]
    assert "-p" in calls[0]


@pytest.mark.basic
def test_default_process_specs_pin_uat_env_defaults(tmp_path: Path) -> None:
    from abstractgateway.maintenance.process_manager import default_process_specs

    specs = default_process_specs(repo_root=tmp_path)
    for pid, expected in [
        ("gateway_uat", {"ABSTRACTGATEWAY_UAT_PORT": "6081", "ABSTRACTGATEWAY_UAT_DATA_DIR": "runtime/gateway_uat"}),
        ("abstractobserver_uat", {"ABSTRACTOBSERVER_UAT_PORT": "6082"}),
        ("abstractcode_web_uat", {"ABSTRACTCODE_WEB_UAT_PORT": "6083"}),
        ("abstractflow_frontend_uat", {"ABSTRACTFLOW_FRONTEND_UAT_PORT": "6084", "ABSTRACTFLOW_BACKEND_UAT_PORT": "6080"}),
        ("abstractflow_backend_uat", {"ABSTRACTFLOW_BACKEND_UAT_PORT": "6080", "ABSTRACTFLOW_RUNTIME_DIR": "runtime/abstractflow_uat"}),
    ]:
        spec = specs.get(pid)
        assert spec is not None
        env = spec.env or {}
        for k, v in expected.items():
            assert env.get(k) == v




from __future__ import annotations

import pytest
from fastapi.testclient import TestClient


def test_gateway_email_inbox_endpoints_contract(monkeypatch: pytest.MonkeyPatch) -> None:
    token = "t"
    monkeypatch.setenv("ABSTRACTGATEWAY_AUTH_TOKEN", token)
    monkeypatch.setenv("ABSTRACTGATEWAY_ALLOWED_ORIGINS", "*")

    from abstractcore.tools import comms_tools

    monkeypatch.setattr(
        comms_tools,
        "list_email_accounts",
        lambda: {
            "success": True,
            "source": "yaml",
            "config_path": "configs/emails.yaml",
            "default_account": "default",
            "accounts": [
                {
                    "account": "default",
                    "email": "alpha@example.com",
                    "from_email": "alpha@example.com",
                    "can_read": True,
                    "can_send": True,
                    "imap_password_set": True,
                    "smtp_password_set": True,
                },
                {
                    "account": "contact",
                    "email": "contact@example.com",
                    "from_email": "contact@example.com",
                    "can_read": True,
                    "can_send": True,
                    "imap_password_set": True,
                    "smtp_password_set": True,
                },
            ],
        },
    )

    monkeypatch.setattr(
        comms_tools,
        "list_emails",
        lambda **kwargs: {
            "success": True,
            "account": kwargs.get("account") or "default",
            "mailbox": kwargs.get("mailbox") or "INBOX",
            "filter": {
                "since": "2026-02-04T00:00:00+00:00",
                "status": kwargs.get("status") or "all",
                "limit": kwargs.get("limit") or 20,
            },
            "counts": {"returned": 1, "unread": 1, "read": 0},
            "messages": [
                {
                    "uid": "123",
                    "message_id": "<m1@example.com>",
                    "subject": "Hello",
                    "from": "sender@example.com",
                    "to": "alpha@example.com",
                    "date": "Tue, 04 Feb 2026 00:00:00 +0000",
                    "flags": ["\\Seen"],
                    "seen": False,
                    "size": 42,
                }
            ],
        },
    )

    monkeypatch.setattr(
        comms_tools,
        "read_email",
        lambda **kwargs: {
            "success": True,
            "account": kwargs.get("account") or "default",
            "mailbox": kwargs.get("mailbox") or "INBOX",
            "uid": kwargs.get("uid") or "",
            "message_id": "<m1@example.com>",
            "subject": "Hello",
            "from": "sender@example.com",
            "to": "alpha@example.com",
            "cc": "",
            "date": "Tue, 04 Feb 2026 00:00:00 +0000",
            "flags": ["\\Seen"],
            "seen": True,
            "body_text": "Hi",
            "body_html": "<p>Hi</p>",
            "attachments": [{"filename": "a.txt", "content_type": "text/plain"}],
        },
    )

    monkeypatch.setattr(
        comms_tools,
        "send_email",
        lambda **kwargs: {
            "success": True,
            "account": kwargs.get("account") or "default",
            "message_id": "<sent@example.com>",
            "from": "alpha@example.com",
            "to": ["to@example.com"],
            "cc": [],
            "bcc": [],
            "smtp": {"host": "smtp.example.com", "port": 587, "username": "alpha@example.com", "starttls": True},
        },
    )

    from abstractgateway.app import app

    client = TestClient(app)
    headers = {"Authorization": f"Bearer {token}"}

    resp = client.get("/api/gateway/email/accounts", headers=headers)
    assert resp.status_code == 200, resp.text
    payload = resp.json()
    assert payload["ok"] is True
    assert payload["default_account"] == "default"
    assert {a["account"] for a in payload["accounts"]} == {"default", "contact"}

    resp2 = client.get("/api/gateway/email/messages", params={"account": "default", "since": "1d", "status": "unread", "limit": 5}, headers=headers)
    assert resp2.status_code == 200, resp2.text
    payload2 = resp2.json()
    assert payload2["ok"] is True
    assert payload2["account"] == "default"
    assert payload2["messages"][0]["uid"] == "123"
    assert payload2["messages"][0]["from"] == "sender@example.com"

    resp3 = client.get("/api/gateway/email/messages/123", params={"account": "default", "max_body_chars": 5000}, headers=headers)
    assert resp3.status_code == 200, resp3.text
    payload3 = resp3.json()
    assert payload3["ok"] is True
    assert payload3["uid"] == "123"
    assert payload3["from"] == "sender@example.com"
    assert payload3["attachments"][0]["filename"] == "a.txt"

    resp4 = client.post(
        "/api/gateway/email/send",
        json={"account": "default", "to": "to@example.com", "subject": "Hi", "body_text": "Hi"},
        headers=headers,
    )
    assert resp4.status_code == 200, resp4.text
    payload4 = resp4.json()
    assert payload4["ok"] is True
    assert payload4["account"] == "default"
    assert payload4["from"] == "alpha@example.com"
    assert payload4["to"] == ["to@example.com"]




from __future__ import annotations

import json
import re
from pathlib import Path
from typing import Any, Dict, Optional, Tuple

from .report_models import ReportHeader, ReportRecord, ReportType


_TITLE_RE = re.compile(r"^#\s+(?P<kind>Bug|Feature)\s*:\s*(?P<title>.+?)\s*$", re.IGNORECASE)
_BLOCKQUOTE_KV_RE = re.compile(r"^\s*>\s*(?P<key>[^:]+?)\s*:\s*(?P<value>.*?)\s*$")
_H2_RE = re.compile(r"^##\s+(?P<name>.+?)\s*$")
_FENCE_JSON_RE = re.compile(r"```json\s*(?P<body>.*?)\s*```", re.IGNORECASE | re.DOTALL)


def _infer_report_type_from_path(path: Path) -> Optional[ReportType]:
    parts = {p.lower() for p in path.parts}
    if "bug_reports" in parts:
        return "bug"
    if "feature_requests" in parts:
        return "feature"
    return None


def _parse_title_and_type(text: str, *, path: Path) -> Tuple[ReportType, str]:
    for raw in text.splitlines():
        line = raw.strip()
        if not line.startswith("#"):
            continue
        m = _TITLE_RE.match(line)
        if m:
            kind = (m.group("kind") or "").strip().lower()
            title = (m.group("title") or "").strip()
            if len(title) > 120:
                title = title[:120].rstrip()
            return ("bug" if kind == "bug" else "feature", title)
        break
    inferred = _infer_report_type_from_path(path) or "bug"
    # Best-effort: use first non-empty line as title.
    for raw in text.splitlines():
        s = raw.strip().lstrip("#").strip()
        if s:
            return inferred, s[:120]
    return inferred, "Report"


def _parse_blockquote_headers(text: str) -> Dict[str, str]:
    out: Dict[str, str] = {}
    for raw in text.splitlines():
        m = _BLOCKQUOTE_KV_RE.match(raw)
        if not m:
            continue
        key = str(m.group("key") or "").strip().lower()
        val = str(m.group("value") or "").strip()
        if key:
            out[key] = val
    return out


def _parse_sections(text: str) -> Dict[str, str]:
    sections: Dict[str, list[str]] = {}
    current: Optional[str] = None
    for raw in text.splitlines():
        m = _H2_RE.match(raw.strip())
        if m:
            current = str(m.group("name") or "").strip()
            if current:
                sections.setdefault(current, [])
            continue
        if current:
            sections[current].append(raw.rstrip("\n"))
    # Preserve leading whitespace (many templates store user-provided fields as
    # Markdown literals via indentation). Only strip surrounding newlines.
    return {k: "\n".join(v).strip("\n") for k, v in sections.items()}


def _extract_context_json(text: str) -> Dict[str, Any]:
    m = _FENCE_JSON_RE.search(text)
    if not m:
        return {}
    body = str(m.group("body") or "").strip()
    if not body:
        return {}
    try:
        obj = json.loads(body)
    except Exception:
        return {}
    return obj if isinstance(obj, dict) else {}


def _first_nonempty_section(sections: Dict[str, str], *names: str) -> str:
    for name in names:
        content = sections.get(name)
        if isinstance(content, str) and content.strip():
            # Preserve indentation (see `_parse_sections`).
            return content.strip("\n")
    return ""


def _dedent_markdown_literal(text: str) -> str:
    """Remove a single Markdown-literal indentation level (4 spaces).

    Gateway report templates store user-provided descriptions as indented code
    blocks to avoid accidental Markdown/HTML rendering. Downstream consumers
    (triage + proposed backlog drafts) want the raw user text.
    """

    s = str(text or "").replace("\r\n", "\n").replace("\r", "\n")
    if not s:
        return ""
    lines = s.split("\n")
    out: list[str] = []
    for ln in lines:
        if ln.startswith("    "):
            out.append(ln[4:])
        elif ln.startswith("\t"):
            out.append(ln[1:])
        else:
            out.append(ln)
    return "\n".join(out).rstrip("\n")


def parse_report_file(path: Path) -> ReportRecord:
    text = path.read_text(encoding="utf-8", errors="replace")
    report_type, title = _parse_title_and_type(text, path=path)

    headers = _parse_blockquote_headers(text)

    def _h(*keys: str) -> str:
        for k in keys:
            v = headers.get(k.lower())
            if isinstance(v, str) and v.strip():
                return v.strip()
        return ""

    created_at = _h("created")
    report_id = _h("bug id", "feature id", "report id")
    session_id = _h("session id")
    session_memory_run_id = _h("session memory run id", "session_memory_run_id")
    active_run_id = _h("relevant run id", "active run id", "run id")
    workflow_id = _h("workflow id")

    sections = _parse_sections(text)

    description = _first_nonempty_section(
        sections,
        "User Description",
        "User Request",
        "Description",
    )
    description = _dedent_markdown_literal(description)

    # Environment info (best-effort; present in templates but may be missing/edited).
    env_section = sections.get("Environment") or ""
    client = ""
    client_version = ""
    provider = ""
    model = ""
    template = ""
    if env_section:
        # Parse "- Key: value" bullets.
        for raw in env_section.splitlines():
            line = raw.strip()
            if not line.startswith("-"):
                continue
            kv = line.lstrip("-").strip()
            if ":" not in kv:
                continue
            k, v = kv.split(":", 1)
            key = k.strip().lower()
            val = v.strip()
            if not val:
                continue
            if key == "client":
                client = val
            elif key == "client version":
                client_version = val
            elif key in {"provider/model", "provider"}:
                provider = val
            elif key == "model":
                model = val
            elif key in {"agent template", "template"}:
                template = val
            elif key == "provider/model":
                # "Provider/model: provider / model"
                if "/" in val:
                    pass

        # Special-case provider/model format: "- Provider/model: X / Y"
        m = re.search(r"^\s*-\s*Provider/model\s*:\s*(.+?)\s*/\s*(.+?)\s*$", env_section, re.IGNORECASE | re.MULTILINE)
        if m:
            provider = (m.group(1) or "").strip()
            model = (m.group(2) or "").strip()

    header = ReportHeader(
        title=title,
        created_at=created_at,
        report_id=report_id,
        session_id=session_id,
        session_memory_run_id=session_memory_run_id,
        active_run_id=active_run_id,
        workflow_id=workflow_id,
        client=client,
        client_version=client_version,
        provider=provider,
        model=model,
        template=template,
    )
    context = _extract_context_json(text)

    return ReportRecord(
        report_type=report_type,
        path=path,
        header=header,
        description=description,
        sections=sections,
        context=context,
    )



from __future__ import annotations

import json
import sys
import time
from pathlib import Path

import pytest


def _wait_until(predicate, *, timeout_s: float = 4.0, poll_s: float = 0.05) -> None:
    end = time.time() + float(timeout_s)
    while time.time() < end:
        if predicate():
            return
        time.sleep(float(poll_s))
    raise AssertionError("timeout waiting for condition")


@pytest.mark.integration
def test_process_manager_env_overrides_apply_before_spec_env(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    base_dir = tmp_path / "runtime"
    repo_root = tmp_path / "repo"
    repo_root.mkdir(parents=True, exist_ok=True)

    # Baseline env value.
    monkeypatch.setenv("ABSTRACT_EMAIL_FROM", "baseline@example.com")

    # Persisted override store.
    state_dir = base_dir / "process_manager"
    state_dir.mkdir(parents=True, exist_ok=True)
    env_path = state_dir / "env_overrides.json"
    env_path.write_text(
        json.dumps(
            {
                "version": 1,
                "updated_at": "now",
                "vars": {
                    "ABSTRACT_EMAIL_FROM": {
                        "enabled": True,
                        "value": "override@example.com",
                        "updated_at": "now",
                    }
                },
            },
            indent=2,
        ),
        encoding="utf-8",
    )

    from abstractgateway.maintenance.process_manager import ProcessManager, ProcessSpec

    # Process that prints the env var then sleeps so we can tail logs.
    cmd = [
        sys.executable,
        "-c",
        "import os, time; print(os.getenv('ABSTRACT_EMAIL_FROM',''), flush=True); time.sleep(60)",
    ]

    override_only = ProcessSpec(id="override_only", label="Override only", kind="service", cwd=".", command=cmd, env={})
    override_only.validate()

    spec_wins = ProcessSpec(
        id="spec_wins",
        label="Spec wins",
        kind="service",
        cwd=".",
        command=cmd,
        env={"ABSTRACT_EMAIL_FROM": "spec@example.com"},
    )
    spec_wins.validate()

    mgr = ProcessManager(base_dir=base_dir, repo_root=repo_root, specs={"override_only": override_only, "spec_wins": spec_wins})

    # override_only should see the override value.
    st = mgr.start("override_only")
    assert st.get("status") == "running"
    try:
        def _has_override() -> bool:
            out = mgr.log_tail("override_only", max_bytes=8000)
            return "override@example.com" in str(out.get("content") or "")

        _wait_until(_has_override)
    finally:
        mgr.stop("override_only")

    # spec_wins should see the spec.env value (wins over overrides).
    st2 = mgr.start("spec_wins")
    assert st2.get("status") == "running"
    try:
        def _has_spec() -> bool:
            out = mgr.log_tail("spec_wins", max_bytes=8000)
            return "spec@example.com" in str(out.get("content") or "")

        _wait_until(_has_spec)
    finally:
        mgr.stop("spec_wins")




from __future__ import annotations

import base64
import hashlib
import hmac
import json
import time
from typing import Any, Dict, Optional, Tuple


def _b64url_encode(data: bytes) -> str:
    return base64.urlsafe_b64encode(data).decode("utf-8").rstrip("=")


def _b64url_decode(text: str) -> bytes:
    s = str(text or "")
    pad = "=" * ((4 - (len(s) % 4)) % 4)
    return base64.urlsafe_b64decode((s + pad).encode("utf-8"))


def sign_action_token(*, payload: Dict[str, Any], secret: str) -> str:
    blob = json.dumps(payload, ensure_ascii=False, sort_keys=True, separators=(",", ":")).encode("utf-8")
    sig = hmac.new(str(secret).encode("utf-8"), blob, hashlib.sha256).digest()
    return f"{_b64url_encode(blob)}.{_b64url_encode(sig)}"


def verify_action_token(*, token: str, secret: str) -> Tuple[Optional[Dict[str, Any]], Optional[str]]:
    raw = str(token or "").strip()
    if not raw or "." not in raw:
        return None, "Invalid token"
    a, b = raw.split(".", 1)
    try:
        blob = _b64url_decode(a)
        sig = _b64url_decode(b)
    except Exception:
        return None, "Invalid token encoding"

    expected = hmac.new(str(secret).encode("utf-8"), blob, hashlib.sha256).digest()
    if not hmac.compare_digest(sig, expected):
        return None, "Invalid token signature"

    try:
        payload = json.loads(blob.decode("utf-8"))
    except Exception:
        return None, "Invalid token payload"
    if not isinstance(payload, dict):
        return None, "Invalid token payload"

    exp = payload.get("exp")
    try:
        exp_i = int(exp) if exp is not None else None
    except Exception:
        exp_i = None
    if exp_i is not None and exp_i > 0:
        now = int(time.time())
        if now > exp_i:
            return None, "Token expired"

    return payload, None


def build_action_payload(*, decision_id: str, action: str, ttl_s: int = 7 * 24 * 3600) -> Dict[str, Any]:
    now = int(time.time())
    exp = now + int(ttl_s) if int(ttl_s) > 0 else 0
    return {"decision_id": str(decision_id), "action": str(action), "iat": now, "exp": exp}


def build_action_links(
    *,
    decision_id: str,
    base_url: str,
    secret: str,
    ttl_s: int = 7 * 24 * 3600,
) -> Dict[str, str]:
    """Build triage action URLs for a decision.

    Returns a dict with keys: approve, defer_1d, defer_7d, reject.
    """
    base = str(base_url or "").rstrip("/")
    if not base:
        return {}

    def _url(payload: Dict[str, Any]) -> str:
        token = sign_action_token(payload=payload, secret=secret)
        return f"{base}/api/triage/action/{token}"

    approve = _url(build_action_payload(decision_id=decision_id, action="approve", ttl_s=ttl_s))
    reject = _url(build_action_payload(decision_id=decision_id, action="reject", ttl_s=ttl_s))

    defer1 = build_action_payload(decision_id=decision_id, action="defer", ttl_s=ttl_s)
    defer1["days"] = 1
    defer7 = build_action_payload(decision_id=decision_id, action="defer", ttl_s=ttl_s)
    defer7["days"] = 7

    return {
        "approve": approve,
        "defer_1d": _url(defer1),
        "defer_7d": _url(defer7),
        "reject": reject,
    }



from __future__ import annotations

import importlib.util
import sys
import time
from pathlib import Path

import pytest
from fastapi.testclient import TestClient


@pytest.fixture(autouse=True)
def _cleanup_optional_abstractflow_modules_after_test():
    yield
    for name in list(sys.modules):
        if name == "abstractflow" or name.startswith("abstractflow."):
            sys.modules.pop(name, None)


def _editor_flow_payload() -> dict:
    return {
        "name": "Gateway Editor Contract",
        "description": "Draft flow used by the gateway-first editor contract test.",
        "interfaces": ["abstractflow.editor.test.v1"],
        "nodes": [
            {
                "id": "start",
                "type": "on_flow_start",
                "position": {"x": 32.0, "y": 128.0},
                "data": {
                    "nodeType": "on_flow_start",
                    "label": "On Flow Start",
                    "inputs": [],
                    "outputs": [
                        {"id": "exec-out", "label": "", "type": "execution"},
                        {"id": "prompt", "label": "Prompt", "type": "string"},
                        {"id": "max_iterations", "label": "Max Iterations", "type": "integer"},
                    ],
                    "pinDefaults": {"prompt": "hello from the editor", "max_iterations": 3},
                },
            },
            {
                "id": "end",
                "type": "on_flow_end",
                "position": {"x": 320.0, "y": 128.0},
                "data": {
                    "nodeType": "on_flow_end",
                    "label": "On Flow End",
                    "inputs": [{"id": "exec-in", "label": "", "type": "execution"}],
                    "outputs": [],
                },
            },
        ],
        "edges": [
            {
                "id": "edge-start-end",
                "source": "start",
                "sourceHandle": "exec-out",
                "target": "end",
                "targetHandle": "exec-in",
            }
        ],
        "entryNode": "start",
    }


def _wait_until(fn, *, timeout_s: float = 5.0, poll_s: float = 0.05) -> None:
    deadline = time.monotonic() + timeout_s
    last_error: AssertionError | None = None
    while time.monotonic() < deadline:
        try:
            if fn():
                return
        except AssertionError as e:
            last_error = e
        time.sleep(poll_s)
    if last_error is not None:
        raise last_error
    raise AssertionError("condition did not become true before timeout")


@pytest.mark.basic
def test_abstractflow_gateway_first_editor_contract_path(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    if importlib.util.find_spec("abstractflow") is None:
        pytest.skip("abstractflow is not installed")

    runtime_dir = tmp_path / "runtime"
    bundles_dir = tmp_path / "bundles"
    token = "t"
    monkeypatch.setenv("ABSTRACTGATEWAY_DATA_DIR", str(runtime_dir))
    monkeypatch.setenv("ABSTRACTGATEWAY_FLOWS_DIR", str(bundles_dir))
    monkeypatch.setenv("ABSTRACTGATEWAY_WORKFLOW_SOURCE", "bundle")
    monkeypatch.setenv("ABSTRACTGATEWAY_AUTH_TOKEN", token)
    monkeypatch.setenv("ABSTRACTGATEWAY_ALLOWED_ORIGINS", "*")
    monkeypatch.setenv("ABSTRACTGATEWAY_POLL_S", "0.05")
    monkeypatch.setenv("ABSTRACTGATEWAY_TICK_WORKERS", "1")

    from abstractgateway.app import app

    headers = {"Authorization": f"Bearer {token}"}
    with TestClient(app) as client:
        unauth = client.get("/api/gateway/visualflows")
        assert unauth.status_code in {401, 403}, unauth.text

        create = client.post("/api/gateway/visualflows", json=_editor_flow_payload(), headers=headers)
        assert create.status_code == 200, create.text
        draft = create.json()
        flow_id = str(draft.get("id") or "")
        assert flow_id
        assert draft.get("name") == "Gateway Editor Contract"

        update = client.put(
            f"/api/gateway/visualflows/{flow_id}",
            json={"description": "Updated by the editor contract test."},
            headers=headers,
        )
        assert update.status_code == 200, update.text
        assert update.json().get("description") == "Updated by the editor contract test."

        fetched = client.get(f"/api/gateway/visualflows/{flow_id}", headers=headers)
        assert fetched.status_code == 200, fetched.text
        assert fetched.json().get("id") == flow_id

        listed = client.get("/api/gateway/visualflows", headers=headers)
        assert listed.status_code == 200, listed.text
        assert any(isinstance(item, dict) and item.get("id") == flow_id for item in listed.json())

        publish = client.post(
            f"/api/gateway/visualflows/{flow_id}/publish",
            json={"bundle_id": "editor-contract", "bundle_version": "0.0.0", "overwrite": True, "reload_gateway": True},
            headers=headers,
        )
        assert publish.status_code == 200, publish.text
        published = publish.json()
        assert published.get("ok") is True
        assert published.get("bundle_id") == "editor-contract"
        assert published.get("bundle_version") == "0.0.0"

        caps = client.get("/api/gateway/discovery/capabilities", headers=headers)
        assert caps.status_code == 200, caps.text
        flow_editor = caps.json()["capabilities"]["contracts"]["flow_editor"]
        assert flow_editor["visualflows"]["crud"]["available"] is True
        assert flow_editor["visualflows"]["publish"]["available"] is True
        assert flow_editor["run_input_schema"]["endpoint"] == "/api/gateway/bundles/{bundle_id}/flows/{flow_id}/input_schema"

        schema = client.get(
            f"/api/gateway/bundles/editor-contract/flows/{flow_id}/input_schema?bundle_version=0.0.0",
            headers=headers,
        )
        assert schema.status_code == 200, schema.text
        schema_body = schema.json()
        assert schema_body.get("version") == 1
        assert schema_body.get("workflow_id") == f"editor-contract@0.0.0:{flow_id}"
        inputs = schema_body.get("inputs")
        assert isinstance(inputs, list)
        assert {item.get("id") for item in inputs if isinstance(item, dict)} == {"prompt", "max_iterations"}
        assert schema_body.get("defaults") == {"prompt": "hello from the editor", "max_iterations": 3}
        props = schema_body.get("input_data_schema", {}).get("properties")
        assert isinstance(props, dict)
        assert props.get("prompt", {}).get("type") == "string"
        assert props.get("max_iterations", {}).get("type") == "integer"

        start = client.post(
            "/api/gateway/runs/start",
            json={
                "bundle_id": "editor-contract",
                "bundle_version": "0.0.0",
                "flow_id": flow_id,
                "input_data": {"prompt": "run from editor", "max_iterations": 5, "extra": "ignored"},
            },
            headers=headers,
        )
        assert start.status_code == 200, start.text
        run_id = start.json()["run_id"]

        def _completed() -> bool:
            run = client.get(f"/api/gateway/runs/{run_id}", headers=headers)
            assert run.status_code == 200, run.text
            return run.json().get("status") == "completed"

        _wait_until(_completed, timeout_s=10.0)

        ledger = client.get(f"/api/gateway/runs/{run_id}/ledger?after=0&limit=200", headers=headers)
        assert ledger.status_code == 200, ledger.text
        ledger_body = ledger.json()
        assert isinstance(ledger_body.get("items"), list)
        assert ledger_body.get("next_after", 0) >= len(ledger_body["items"])

        batch = client.post(
            "/api/gateway/runs/ledger/batch",
            json={"limit": 200, "runs": [{"run_id": run_id, "after": 0}]},
            headers=headers,
        )
        assert batch.status_code == 200, batch.text
        assert run_id in (batch.json().get("runs") or {})

        input_data = client.get(f"/api/gateway/runs/{run_id}/input_data", headers=headers)
        assert input_data.status_code == 200, input_data.text
        input_body = input_data.json()
        assert input_body.get("bundle_id") == "editor-contract"
        assert input_body.get("bundle_version") == "0.0.0"
        assert input_body.get("flow_id") == flow_id
        assert input_body.get("input_data") == {"prompt": "run from editor", "max_iterations": 5}

        artifacts = client.get(f"/api/gateway/runs/{run_id}/artifacts", headers=headers)
        assert artifacts.status_code == 200, artifacts.text
        assert isinstance(artifacts.json().get("items"), list)

        history = client.get(
            f"/api/gateway/runs/{run_id}/history_bundle?include_subruns=true&ledger_mode=tail&ledger_max_items=200",
            headers=headers,
        )
        assert history.status_code == 200, history.text
        history_body = history.json()
        assert history_body.get("root_run_id") == run_id
        assert isinstance(history_body.get("run"), dict)
        assert history_body["run"].get("run_id") == run_id

        delete = client.delete(f"/api/gateway/visualflows/{flow_id}", headers=headers)
        assert delete.status_code == 200, delete.text
        assert delete.json().get("status") == "deleted"


@pytest.mark.basic
def test_abstractflow_gateway_publish_fails_fast_when_reload_fails(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    if importlib.util.find_spec("abstractflow") is None:
        pytest.skip("abstractflow is not installed")

    runtime_dir = tmp_path / "runtime"
    monkeypatch.setenv("ABSTRACTGATEWAY_DATA_DIR", str(runtime_dir))
    monkeypatch.setenv("ABSTRACTGATEWAY_FLOWS_DIR", str(tmp_path / "flows"))
    monkeypatch.setenv("ABSTRACTGATEWAY_WORKFLOW_SOURCE", "bundle")
    monkeypatch.setenv("ABSTRACTGATEWAY_AUTH_TOKEN", "t")
    monkeypatch.setenv("ABSTRACTGATEWAY_ALLOWED_ORIGINS", "*")
    monkeypatch.setenv("ABSTRACTGATEWAY_POLL_S", "0.05")
    monkeypatch.setenv("ABSTRACTGATEWAY_TICK_WORKERS", "1")

    from abstractgateway.app import app
    from abstractgateway.service import get_gateway_service

    headers = {"Authorization": "Bearer t"}
    with TestClient(app) as client:
        create = client.post("/api/gateway/visualflows", json=_editor_flow_payload(), headers=headers)
        assert create.status_code == 200, create.text
        flow_id = str(create.json().get("id") or "")
        assert flow_id

        svc = get_gateway_service()
        host = getattr(svc, "host", None)
        assert host is not None

        def _broken_reload() -> None:
            raise RuntimeError("simulated bundle reload failure")

        monkeypatch.setattr(host, "reload_bundles_from_disk", _broken_reload)

        publish = client.post(
            f"/api/gateway/visualflows/{flow_id}/publish",
            json={"bundle_id": "editor-contract", "bundle_version": "0.0.1", "overwrite": True, "reload_gateway": True},
            headers=headers,
        )
        assert publish.status_code == 503, publish.text
        assert "Failed to reload bundles after publish" in publish.text



from __future__ import annotations

import hashlib
from dataclasses import dataclass
from pathlib import Path

import pytest
from fastapi import FastAPI
from fastapi.testclient import TestClient


@dataclass
class _Stores:
    base_dir: Path


@dataclass
class _Service:
    stores: _Stores


def _make_app(*, monkeypatch: pytest.MonkeyPatch, gateway_base_dir: Path) -> FastAPI:
    import abstractgateway.routes.gateway as gateway_routes

    monkeypatch.setattr(gateway_routes, "get_gateway_service", lambda: _Service(stores=_Stores(base_dir=gateway_base_dir)))

    app = FastAPI()
    app.include_router(gateway_routes.router, prefix="/api")
    return app


def _write_backlog_template(path: Path) -> None:
    path.write_text(
        "\n".join(
            [
                "# {ID}-{Package}: {Title}",
                "",
                "> Created: YYYY-MM-DD HH:MM:SS +/-TZ",
                "",
                "## Summary",
                "One paragraph describing what this task accomplishes (user value + outcome).",
                "",
                "## Related",
                "- ADRs:",
                "- Code:",
                "- Reports:",
                "",
            ]
        )
        + "\n",
        encoding="utf-8",
    )


@pytest.mark.basic
def test_backlog_template_and_attachments(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    gateway_dir = tmp_path / "gateway"
    gateway_dir.mkdir(parents=True, exist_ok=True)

    repo_root = tmp_path / "repo"
    backlog_root = repo_root / "docs" / "backlog"
    for d in ["planned", "completed", "proposed", "recurrent", "deprecated", "trash"]:
        (backlog_root / d).mkdir(parents=True, exist_ok=True)
    _write_backlog_template(backlog_root / "template.md")

    monkeypatch.setenv("ABSTRACTGATEWAY_TRIAGE_REPO_ROOT", str(repo_root))

    app = _make_app(monkeypatch=monkeypatch, gateway_base_dir=gateway_dir)
    with TestClient(app) as client:
        templ = client.get("/api/gateway/backlog/template")
        assert templ.status_code == 200
        body = templ.json()
        assert body["relpath"] == "docs/backlog/template.md"
        assert "## Summary" in body["content"]
        assert isinstance(body["sha256"], str) and len(body["sha256"]) == 64

        created = client.post("/api/gateway/backlog/create", json={"kind": "proposed", "package": "framework", "title": "Example task"})
        assert created.status_code == 200
        created_body = created.json()
        filename = created_body["filename"]
        item_id = int(created_body["item_id"])
        pid = f"{item_id:03d}"

        # Upload attachment (name sanitization + collision suffixing).
        payload = b"hello"
        sha = hashlib.sha256(payload).hexdigest()
        files = {"file": ("screenshot 1.png", payload, "image/png")}
        up = client.post(f"/api/gateway/backlog/proposed/{filename}/attachments/upload", files=files)
        assert up.status_code == 200
        up_body = up.json()
        assert up_body["item_id"] == item_id
        stored = up_body["stored"]
        assert stored["sha256"] == sha
        assert stored["relpath"].startswith(f"docs/backlog/assets/{pid}/")
        assert stored["filename"] == "screenshot_1.png"

        on_disk = repo_root / stored["relpath"]
        assert on_disk.exists()
        assert on_disk.read_bytes() == payload

        # Second upload with same name should suffix.
        up2 = client.post(f"/api/gateway/backlog/proposed/{filename}/attachments/upload", files=files)
        assert up2.status_code == 200
        stored2 = up2.json()["stored"]
        assert stored2["filename"].startswith("screenshot_1-2")
        on_disk2 = repo_root / stored2["relpath"]
        assert on_disk2.exists()




from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path

import pytest
from fastapi import FastAPI
from fastapi.testclient import TestClient


@dataclass
class _Stores:
    base_dir: Path


@dataclass
class _Service:
    stores: _Stores


def _make_app(*, monkeypatch: pytest.MonkeyPatch, gateway_base_dir: Path) -> FastAPI:
    import abstractgateway.routes.gateway as gateway_routes

    monkeypatch.setattr(gateway_routes, "get_gateway_service", lambda: _Service(stores=_Stores(base_dir=gateway_base_dir)))

    app = FastAPI()
    app.include_router(gateway_routes.router, prefix="/api")
    return app


def _write_backlog_item(path: Path, *, item_id: int, pkg: str, title: str) -> None:
    path.write_text(
        "\n".join(
            [
                f"# {item_id}-{pkg}: {title}",
                "",
                "> Created: 2026-01-31 00:00:00 +0000",
                "",
                "## Summary",
                "Test item.",
                "",
            ]
        )
        + "\n",
        encoding="utf-8",
    )


@pytest.mark.basic
def test_backlog_endpoints_support_recurrent_and_unparsed(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    gateway_dir = tmp_path / "gateway"
    gateway_dir.mkdir(parents=True, exist_ok=True)

    repo_root = tmp_path / "repo"
    backlog_root = repo_root / "docs" / "backlog"
    (backlog_root / "planned").mkdir(parents=True, exist_ok=True)
    (backlog_root / "completed").mkdir(parents=True, exist_ok=True)
    (backlog_root / "proposed").mkdir(parents=True, exist_ok=True)
    (backlog_root / "recurrent").mkdir(parents=True, exist_ok=True)

    _write_backlog_item(backlog_root / "recurrent" / "700-framework-example-recurrent.md", item_id=700, pkg="framework", title="Example recurrent")
    # Write an unparsed backlog file (doesn't match required # {id}-{pkg}: {title} format).
    (backlog_root / "planned" / "unparsed.md").write_text("hello\n", encoding="utf-8")

    monkeypatch.setenv("ABSTRACTGATEWAY_TRIAGE_REPO_ROOT", str(repo_root))

    app = _make_app(monkeypatch=monkeypatch, gateway_base_dir=gateway_dir)
    with TestClient(app) as client:
        recurrent = client.get("/api/gateway/backlog/recurrent")
        assert recurrent.status_code == 200
        items = recurrent.json()["items"]
        assert any(i["filename"] == "700-framework-example-recurrent.md" and i["item_id"] == 700 for i in items)

        planned = client.get("/api/gateway/backlog/planned")
        assert planned.status_code == 200
        pitems = planned.json()["items"]
        unparsed = next((i for i in pitems if i["filename"] == "unparsed.md"), None)
        assert unparsed is not None
        assert unparsed["item_id"] == 0
        assert unparsed.get("parsed") is False

        content = client.get("/api/gateway/backlog/recurrent/700-framework-example-recurrent.md/content")
        assert content.status_code == 200
        assert "# 700-framework: Example recurrent" in content.json()["content"]

        raw = client.get("/api/gateway/backlog/planned/unparsed.md/content")
        assert raw.status_code == 200
        assert raw.json()["content"].strip() == "hello"




"""AbstractGateway.

AbstractGateway is a deployable Run Gateway host for AbstractRuntime:
- durable command inbox (start/resume/pause/cancel/emit_event)
- ledger replay + SSE streaming (replay-first)
- security middleware for network-safe deployments
"""

__version__ = "0.2.14"



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict

from abstractgateway.runner import GatewayRunner
from abstractruntime import Effect, EffectType, Runtime, StepPlan, WorkflowSpec
from abstractruntime.core.models import RunStatus, WaitReason
from abstractruntime.integrations.abstractcore.effect_handlers import build_effect_handlers
from abstractruntime.storage.in_memory import InMemoryLedgerStore, InMemoryRunStore


class _StubLLM:
    def generate(self, **kwargs: Any) -> Dict[str, Any]:
        return {"content": "ok", "tool_calls": []}


class _StubTools:
    def execute(self, *, tool_calls: Any) -> Dict[str, Any]:
        return {"mode": "executed", "results": []}


@dataclass
class _Host:
    runtime: Runtime
    registry: Dict[str, WorkflowSpec]
    run_store: InMemoryRunStore
    ledger_store: InMemoryLedgerStore
    artifact_store: Any = None

    def runtime_and_workflow_for_run(self, run_id: str) -> tuple[Runtime, WorkflowSpec]:
        run = self.run_store.load(str(run_id))
        if run is None:
            raise KeyError(f"Run '{run_id}' not found")
        spec = self.registry.get(str(run.workflow_id))
        if spec is None:
            raise KeyError(f"Workflow '{run.workflow_id}' not registered")
        return self.runtime, spec


def test_gateway_runner_resumes_subworkflow_parent_with_node_traces(tmp_path: Path) -> None:
    run_store = InMemoryRunStore()
    ledger_store = InMemoryLedgerStore()

    runtime = Runtime(
        run_store=run_store,
        ledger_store=ledger_store,
        effect_handlers=build_effect_handlers(llm=_StubLLM(), tools=_StubTools()),
    )

    def child_node(run, ctx) -> StepPlan:
        return StepPlan(
            node_id="child_node",
            effect=Effect(type=EffectType.LLM_CALL, payload={"prompt": "hi"}, result_key="_temp.child"),
            next_node=None,
        )

    child = WorkflowSpec(workflow_id="child", entry_node="child_node", nodes={"child_node": child_node})

    def start(run, ctx) -> StepPlan:
        return StepPlan(
            node_id="start",
            effect=Effect(
                type=EffectType.START_SUBWORKFLOW,
                payload={
                    "workflow_id": "child",
                    "vars": {},
                    "async": True,
                    "wait": True,
                    "include_traces": True,
                },
                result_key="sub",
            ),
            next_node="done",
        )

    def done(run, ctx) -> StepPlan:
        return StepPlan(node_id="done", complete_output={"sub": run.vars.get("sub")})

    parent = WorkflowSpec(workflow_id="parent", entry_node="start", nodes={"start": start, "done": done})

    registry: Dict[str, WorkflowSpec] = {"parent": parent, "child": child}
    runtime.set_workflow_registry(registry)

    parent_run_id = runtime.start(workflow=parent, vars={})
    parent_state = runtime.tick(workflow=parent, run_id=parent_run_id, max_steps=5)

    assert parent_state.status == RunStatus.WAITING
    assert parent_state.waiting is not None
    assert parent_state.waiting.reason == WaitReason.SUBWORKFLOW
    assert parent_state.waiting.details.get("include_traces") is True

    child_run_id = parent_state.waiting.details.get("sub_run_id")
    assert isinstance(child_run_id, str) and child_run_id

    host = _Host(runtime=runtime, registry=registry, run_store=run_store, ledger_store=ledger_store)
    runner = GatewayRunner(base_dir=tmp_path, host=host)

    # Drive the child to completion and ensure the parent is resumed with traces.
    runner._tick_run(child_run_id)

    resumed_parent = runtime.get_state(parent_run_id)
    assert resumed_parent.status == RunStatus.RUNNING

    sub = resumed_parent.vars.get("sub")
    assert isinstance(sub, dict)
    assert sub.get("sub_run_id") == child_run_id

    traces = sub.get("node_traces")
    assert isinstance(traces, dict)
    assert traces.get("child_node", {}).get("steps"), "expected child node traces to be propagated"


def test_gateway_runner_marks_child_failed_and_resumes_parent(tmp_path: Path) -> None:
    run_store = InMemoryRunStore()
    ledger_store = InMemoryLedgerStore()

    runtime = Runtime(
        run_store=run_store,
        ledger_store=ledger_store,
        effect_handlers=build_effect_handlers(llm=_StubLLM(), tools=_StubTools()),
    )

    def child_node(_run, _ctx) -> StepPlan:
        raise RuntimeError("boom")

    child = WorkflowSpec(workflow_id="child_fail", entry_node="child_node", nodes={"child_node": child_node})

    def start(run, ctx) -> StepPlan:
        return StepPlan(
            node_id="start",
            effect=Effect(
                type=EffectType.START_SUBWORKFLOW,
                payload={
                    "workflow_id": "child_fail",
                    "vars": {},
                    "async": True,
                    "wait": True,
                },
                result_key="sub",
            ),
            next_node="done",
        )

    def done(run, ctx) -> StepPlan:
        return StepPlan(node_id="done", complete_output={"sub": run.vars.get("sub")})

    parent = WorkflowSpec(workflow_id="parent_fail", entry_node="start", nodes={"start": start, "done": done})

    registry: Dict[str, WorkflowSpec] = {"parent_fail": parent, "child_fail": child}
    runtime.set_workflow_registry(registry)

    parent_run_id = runtime.start(workflow=parent, vars={})
    parent_state = runtime.tick(workflow=parent, run_id=parent_run_id, max_steps=5)

    assert parent_state.status == RunStatus.WAITING
    assert parent_state.waiting is not None
    assert parent_state.waiting.reason == WaitReason.SUBWORKFLOW

    child_run_id = parent_state.waiting.details.get("sub_run_id")
    assert isinstance(child_run_id, str) and child_run_id

    host = _Host(runtime=runtime, registry=registry, run_store=run_store, ledger_store=ledger_store)
    runner = GatewayRunner(base_dir=tmp_path, host=host)

    runner._tick_run(child_run_id)

    child_state = runtime.get_state(child_run_id)
    assert child_state.status == RunStatus.FAILED
    assert "boom" in str(child_state.error or "")

    resumed_parent = runtime.get_state(parent_run_id)
    assert resumed_parent.status == RunStatus.RUNNING

    sub = resumed_parent.vars.get("sub")
    assert isinstance(sub, dict)
    assert sub.get("sub_run_id") == child_run_id
    output = sub.get("output")
    assert isinstance(output, dict)
    assert output.get("success") is False


def test_gateway_runner_recovers_stuck_subworkflow_wait_when_child_already_terminal(tmp_path: Path) -> None:
    run_store = InMemoryRunStore()
    ledger_store = InMemoryLedgerStore()

    runtime = Runtime(
        run_store=run_store,
        ledger_store=ledger_store,
        effect_handlers=build_effect_handlers(llm=_StubLLM(), tools=_StubTools()),
    )

    def child_node(_run, _ctx) -> StepPlan:
        return StepPlan(node_id="child_node", complete_output={"ok": True})

    child = WorkflowSpec(workflow_id="child_terminal", entry_node="child_node", nodes={"child_node": child_node})

    def start(run, _ctx) -> StepPlan:
        return StepPlan(
            node_id="start",
            effect=Effect(
                type=EffectType.START_SUBWORKFLOW,
                payload={"workflow_id": "child_terminal", "vars": {}, "async": True, "wait": True},
                result_key="sub",
            ),
            next_node="done",
        )

    def done(run, _ctx) -> StepPlan:
        return StepPlan(node_id="done", complete_output={"sub": run.vars.get("sub")})

    parent = WorkflowSpec(workflow_id="parent_terminal", entry_node="start", nodes={"start": start, "done": done})

    registry: Dict[str, WorkflowSpec] = {"parent_terminal": parent, "child_terminal": child}
    runtime.set_workflow_registry(registry)

    parent_run_id = runtime.start(workflow=parent, vars={}, actor_id="gateway")
    parent_state = runtime.tick(workflow=parent, run_id=parent_run_id, max_steps=5)

    assert parent_state.status == RunStatus.WAITING
    assert parent_state.waiting is not None
    assert parent_state.waiting.reason == WaitReason.SUBWORKFLOW

    child_run_id = parent_state.waiting.details.get("sub_run_id")
    assert isinstance(child_run_id, str) and child_run_id

    # Simulate a process crash/restart where the child has already reached a terminal state
    # but the runner never got a chance to tick the child and resume the parent.
    child_state = runtime.get_state(child_run_id)
    child_state.status = RunStatus.FAILED
    child_state.error = "boom"
    child_state.updated_at = "2026-01-01T00:00:00+00:00"
    run_store.save(child_state)

    host = _Host(runtime=runtime, registry=registry, run_store=run_store, ledger_store=ledger_store)
    runner = GatewayRunner(base_dir=tmp_path, host=host)

    runner._repair_terminal_subworkflow_waits()

    resumed_parent = runtime.get_state(parent_run_id)
    assert resumed_parent.status == RunStatus.RUNNING
